import os
import shutil
import unittest
# Import torch first and then transformers to avoid unresponsive jit.trace.
import torch
import transformers
from datasets import load_dataset
import neural_compressor.adaptor.pytorch as nc_torch
from intel_extension_for_transformers.optimization import BenchmarkConfig
from intel_extension_for_transformers.optimization.benchmark import benchmark


def get_example_inputs(model_name, dataset_name='sst2'):
    tokenizer = transformers.AutoTokenizer.from_pretrained(model_name)
    dataset = load_dataset(dataset_name, split='validation')
    text = dataset[0]['text'] if dataset_name=='lambada' else dataset[0]['sentence']
    example_inputs = tokenizer(text, padding='max_length', max_length=195, return_tensors='pt')
    return example_inputs

class TestBenchmark(unittest.TestCase):
    def test_fp32_model(self):
        config = BenchmarkConfig(
            batch_size=16,
            cores_per_instance=4,
            num_of_instance=-1,
        )
        model_name_or_path = "distilbert-base-uncased-finetuned-sst-2-english"
        if os.path.exists("/tf_dataset2/models/nlp_toolkit/distilbert_base_uncased_sst2"):
            model_name_or_path = "/tf_dataset2/models/nlp_toolkit/distilbert_base_uncased_sst2"
        example_inputs = get_example_inputs(model_name_or_path)
        benchmark(model_name_or_path, config, example_inputs=example_inputs)

    def test_fp32_model_ipex(self):
        config = BenchmarkConfig(
            backend='ipex',
            batch_size=16,
            cores_per_instance=4,
            num_of_instance=-1,
        )
        model_name_or_path = "distilbert-base-uncased-finetuned-sst-2-english"
        if os.path.exists("/tf_dataset2/models/nlp_toolkit/distilbert_base_uncased_sst2"):
            model_name_or_path = "/tf_dataset2/models/nlp_toolkit/distilbert_base_uncased_sst2"
        example_inputs = get_example_inputs(model_name_or_path)
        benchmark(model_name_or_path, config, example_inputs=example_inputs)

    def test_int8_model(self):
        config = BenchmarkConfig(
            batch_size=16,
            cores_per_instance=4,
            num_of_instance=-1,
        )
        model_name_or_path = "Intel/distilbert-base-uncased-finetuned-sst-2-english-int8-static"
        if os.path.exists("/tf_dataset2/models/nlp_toolkit/distilbert_sst2_int8"):
            model_name_or_path = "/tf_dataset2/models/nlp_toolkit/distilbert_sst2_int8"
        example_inputs = get_example_inputs(model_name_or_path)
        example_inputs = example_inputs['input_ids'] # for UT coverage
        benchmark(model_name_or_path, config, example_inputs=example_inputs)

    def test_int8_model_ipex(self):
        config = BenchmarkConfig(
            backend='ipex',
            batch_size=16,
            cores_per_instance=4,
            num_of_instance=-1,
        )
        model_name_or_path = "Intel/distilbert-base-uncased-finetuned-sst-2-english-int8-static"
        if os.path.exists("/tf_dataset2/models/nlp_toolkit/distilbert_sst2_int8"):
            model_name_or_path = "/tf_dataset2/models/nlp_toolkit/distilbert_sst2_int8"
        example_inputs = get_example_inputs(model_name_or_path)
        example_inputs = tuple(example_inputs.values()) # for UT coverage
        benchmark(model_name_or_path, config, example_inputs=example_inputs)

    def test_torchscript(self):
        config = BenchmarkConfig(
            batch_size=16,
            cores_per_instance=4,
            num_of_instance=-1,
            torchscript=True,
        )
        model_name_or_path = "distilbert-base-uncased-finetuned-sst-2-english"
        if os.path.exists("/tf_dataset2/models/nlp_toolkit/distilbert_base_uncased_sst2"):
            model_name_or_path = "/tf_dataset2/models/nlp_toolkit/distilbert_base_uncased_sst2"
        example_inputs = get_example_inputs(model_name_or_path)
        benchmark(model_name_or_path, config, example_inputs=example_inputs)

    def test_int8_torchscript(self):
        config = BenchmarkConfig(
            batch_size=16,
            cores_per_instance=4,
            num_of_instance=-1,
            torchscript=True,
        )
        model_name_or_path = "Intel/distilbert-base-uncased-finetuned-sst-2-english-int8-static"
        if os.path.exists("/tf_dataset2/models/nlp_toolkit/distilbert_sst2_int8"):
            model_name_or_path = "/tf_dataset2/models/nlp_toolkit/distilbert_sst2_int8"
        example_inputs = get_example_inputs(model_name_or_path)
        benchmark(model_name_or_path, config, example_inputs=example_inputs)

    def test_generate(self):
        config = BenchmarkConfig(
            batch_size=16,
            cores_per_instance=4,
            num_of_instance=1,
            torchscript=False,
            generate=True,
            max_length=195,
        )
        model_name = "adasnew/t5-small-xsum"
        if os.path.exists("/tf_dataset2/models/nlp_toolkit/t5-small-xsum"):
            model_name = "/tf_dataset2/models/nlp_toolkit/t5-small-xsum"
        example_inputs = get_example_inputs(model_name, dataset_name='lambada')
        example_inputs = example_inputs['input_ids'][0].to('cpu').unsqueeze(0) # for UT coverage
        benchmark(model_name, config, example_inputs=example_inputs)

    def test_int8_generate(self):
        config = BenchmarkConfig(
            batch_size=16,
            cores_per_instance=4,
            num_of_instance=1,
            torchscript=False,
            generate=True,
            max_length=195,
        )
        model_name = "Intel/t5-small-xsum-int8-dynamic"
        if os.path.exists("/tf_dataset2/models/nlp_toolkit/t5-small-xsum-int8-dynamic"):
            model_name = "/tf_dataset2/models/nlp_toolkit/t5-small-xsum-int8-dynamic"
        example_inputs = get_example_inputs(model_name, dataset_name='lambada')
        benchmark(model_name, config, example_inputs=example_inputs)

    def test_torchscript_generate(self):
        # test_int8_torchscript_generate will fail when torch.jit.trace, deleted.
        config = BenchmarkConfig(
            batch_size=16,
            cores_per_instance=4,
            num_of_instance=1,
            torchscript=True,
            generate=True,
            max_length=195,
        )
        model_name = "adasnew/t5-small-xsum"
        if os.path.exists("/tf_dataset2/models/nlp_toolkit/t5-small-xsum"):
            model_name = "/tf_dataset2/models/nlp_toolkit/t5-small-xsum"
        example_inputs = get_example_inputs(model_name, dataset_name='lambada')
        benchmark(model_name, config, example_inputs=example_inputs)

if __name__ == "__main__":
    unittest.main()



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The neural engine graph utils."""

from . import logger
import copy
import re
import os
import numpy as np
from collections import namedtuple, OrderedDict
from schema import Schema, And, Or
import importlib

DTYPES_DICT = {
    "float16": "fp16",
    "float32": "fp32",
    "int8": "s8",
    "int32": "s32",
    "int64": "s64",
    "uint8": "u8",
    "uint16": "bf16",
    "float64": "fp64"
}


def names_from_input(name):
    """Static method that get the valid node / tensor name from input name.

    Args:
        name (string): name defined in the input field.

    Returns:
        string tuple: (node's name, tensor's name)

    for example: In NodeDef.input, the name from list is tensor name, may not end with ':0',
                which can not be used for tensor name in the new Graph class. If it end with ':0',
                it can not also be used for node name in the new Graph class
    """
    if name.startswith("^"):
        name = name[1:]
    m = re.search(r"(.*)(:\d+)$", name)
    # not end with ':x"
    if m is None:
        node_name = name
        tensor_name = name + ':0'
    # end with ':x'
    else:
        node_name = m.group(1)
        tensor_name = name

    return (node_name, tensor_name)


def get_data_dtype(data):
    """Get the const data dtype.

    Args:
       data (numpy data): a const data to model

    Returns:
       dtype (String): the value in DTYPES_DICT
    """
    dtype = None
    if np.issubdtype(data.dtype, np.integer) or np.issubdtype(data.dtype, np.floating):
        try:
            dtype = DTYPES_DICT[str(data.dtype[0])]
        except BaseException:
            dtype = DTYPES_DICT[str(data.dtype)]

    return dtype


def autocast_init():
    """Initialize the quant info."""
    global _autocast_info
    _autocast_info = {'cast_type': 'native'}


def set_autocast(key, value):
    """Modify the quant info."""
    _autocast_info[key] = value


def get_autocast_info():
    """Get the quant info."""
    return _autocast_info


def quant_info_init():
    """Initialize the quant info."""
    global _quant_info
    _quant_info = {}


def insert_quant_info(key, value):
    """Modify the quant info."""
    _quant_info[key] = value


def get_quant_info():
    """Get the quant info."""
    return _quant_info


def search_straight_pattern(input_pattern, graph):
    """Search user specified patterns on internal grpah structure.

    Attention: the input computation chain in the graph which can be called pattern, there must be
                straight (or sequence). It means it has not any subgraph nodes. Otherwise this
                function returns []

    Args:
        input_pattern (list): Contains the op_type of the nodes in pattern. The element of the
        list could be string/list/tuple, string or list means the specified op_type are mandatory
        while tuple stands for optional.
        For example, a input pattern mybe like this:
        ['Mul', 'Mul', ['Add', 'AddV2']] it equals to below patterns:
        'Mul' + 'Mul' + 'Add'
        'Mul' + 'Mul' + 'AddV2'
        graph: Graph Class, the new graph generated from extractor.

    Returns: [string list]. The length is the matched pattern results in the graph, for example,
            the graph has 24 layers and each layer has a 'LayerNorm' pattern, then the length is
            24. Each match pattern result is still a list contains the node names, and the last
            element is the op_type list corresponding to the former node names.

            Here is the return result example::


                [
                    ['Mul' node name,
                    'Mul' node name,
                    'Add' node name,
                    ['Mul', 'Mul', 'Add']],

                    ['Mul' node name,
                    'Mul' node name,
                    'AddV2' node name,
                    ['Mul', 'Mul', 'AddV2']],

                    ...
                ]
    """

    def _validate_input(data, creteria):
        """Validation of input data."""
        if isinstance(creteria, str) and data == creteria:
            return True

        if isinstance(creteria, (list, tuple)) and data in creteria:
            return True

        return False

    def _compare_list(list_a, list_b):
        """Check list a is a subset of list b.

        e.g, list a is ['a', 'b', 'c'] while list b is ['a', 'b', 'c', 'd'],
        then list a is subset of list b.
        Args:
            list_a ([Any]): list A
            list_b ([Any]): list B

        Returns:
            [bool]: list a is a subset of list b or not.
        """
        assert isinstance(list_a, list)
        assert isinstance(list_b, list)
        is_subset = True

        for index, value in enumerate(list_a):
            is_subset &= value == list_b[index]

        return is_subset

    def _dfs(op_names, op_types, node, pattern):
        """Use depth-first algorithm to search for patterns according to node types."""
        if pattern == []:
            return
        start_index = 0
        end_index = len(pattern) - 1
        matched_flag = False
        while start_index <= end_index:
            matched_flag = _validate_input(node.op_type, pattern[end_index])

            if not matched_flag and isinstance(pattern[end_index], tuple):
                end_index -= 1
                continue

            if matched_flag:
                op_names.append(node.name)
                op_types.append(node.op_type)
                break

            return

        if start_index == end_index:
            if matched_flag:
                matched_res = copy.deepcopy(op_names)
                matched_res.reverse()
                op_types_copy = copy.deepcopy(op_types)
                op_types_copy.reverse()
                matched_res.append(op_types_copy)
                if matched_res not in output_result:
                    output_result.append(matched_res)

                op_names.pop()
                op_types.pop()
            return

        for index, value in enumerate(node.input_tensors):
            is_const = False
            is_const = (isinstance(value.data, np.ndarray))
            if not is_const and len(value.source_op) != 0:
                try:
                    cur_node = graph.get_node_by_name(value.source_op[0])
                    _dfs(op_names, op_types, cur_node, pattern[:end_index])
                except BaseException:
                    pass
            if index == len(node.input_tensors) - 1:
                op_names.pop()
                op_types.pop()

    output_result = []

    for v in graph.nodes:
        start_index = len(input_pattern) - 1
        while start_index >= 0:
            find_first_match = _validate_input(v.op_type, input_pattern[start_index])
            if find_first_match:
                break

            if isinstance(input_pattern[start_index], tuple):
                start_index -= 1
                continue

            start_index = -2

        if start_index < 0:
            continue

        visited_op_name = []
        visited_op_types = []

        _dfs(visited_op_name, visited_op_types, v, input_pattern)

    sorted_output = sorted(output_result, key=lambda i: i[-1])

    useless_match_list = []
    for index, value in enumerate(sorted_output):

        if index == len(sorted_output) - 1:
            break

        next_matched_op_names = sorted_output[index + 1][:-1]
        if len(value[:-1]) < len(next_matched_op_names) and \
                _compare_list(value[:-1], next_matched_op_names):
            useless_match_list.append(value)

    for i in useless_match_list:
        sorted_output.remove(i)

    longest_match = {}
    final_output = []
    for i in sorted_output:
        key = i[0]
        if key not in longest_match:
            longest_match[key] = i[-1]
            continue

        if len(longest_match[key]) < len(i[-1]):
            longest_match[key] = i[-1]

    for i in sorted_output:
        if i[0] in longest_match and i[-1] == longest_match[i[0]]:
            final_output.append(i)

    return final_output


def search_pattern(pattern_list, graph):
    """Search the complete pattern in the graph.

    Args:
        pattern_list: a list contains  pattern representation. The pattern representation is also
                      a list and each node in the list is a tuple, its form is like "(op_idx,
                      op_type)". However, due to a few complicated patterns, they have sub-graph
                      computation flow. Therefore in a pattern representation, using the fist list
                      represents the main top-down computation flow (from pattern head op to tail
                      op), the left lists represent sub-graphs (their tail nodes must in the main
                      computation flow).

                      Example::

                          #LayerNorm pattern from bert_large_squad.pb
                          [
                              [(0, 'Mean'), (1, 'SquaredDifference'), (2, 'Mean'), (3, 'AddV2'),
                              (4, 'Rsqrt'), (5, 'Mul'), (7 ,'Mul'), (8, 'Sub'), (9, 'AddV2')],
                              [(5, 'Mul'), (6, 'Mul'), (9, 'AddV2')]
                          ]

        graph: Graph Class, the new graph generated from extractor

    Returns: [string list], as same as search_straight_pattern func.

    Note:
        1. The op_idx follows the order in the original frozen model, which means you had better
           not identify them on your own casually.
        2. the main top-down computation flow follows the "tf control flow". It's a straight chain
           from start to end in the pattern. Mostly, it has the longest path. Sometimes, this main
           flow may have sub connection. But you don't need to represent them. A sub-graph must
           have one op at least that doesn't exist in the main chain. For example, in the above
           LayerNorm pattern, (0, 'Mean') has a connection with (7 ,'Mul'). But you don't need to
           represent this relationship.
        3. If a node in sub-graph has several input /output paths, you should split them, each
           sub-graph has one input /output op. (these ops must be in the pattern).
           For example, the below representtaion should be two sub-graphs:
           Add --- Mul --- Sub
           Add ---^
           [..., [(idx, 'Add'),(idx, 'Mul'),(idx, 'Sub')], [(idx, 'Add'),(id, 'Mul'),
            (idx, 'Sub')], ...]
        4. If a node in sub-graph has several input ops, some of them are from outside. Then you
           don't need to give the sub-graphs with the outside op.
           For example, the below representtaion should be one sub-graph:
           Add     ---   Mul --- Sub
           outside op ---^
           [..., [(idx, 'Add'),(idx, 'Mul'),(idx, 'Sub')], ...]
        5. If a node in sub-graph just has one input op and this op is from outside, you should
           use empty tuple () to represents a input op. However, the algorithm doesn't support
           this kind of pattern completely. Beause the match result can't make sure the whole
           connection. So you had better check the results.
        6. For the symmetric pattern, the sub-graph has consecutive same op type as the main chain
           (Y or O shape). So these two search results by DFS are duplicated. The algorithm would
           perform checking before splicing and de-duplication. The sub-graph length <= the main
           chain length.
        7. Some pattern has several same sub-graphs, these sub-graphs have same tail node and op
           types are totally same.

           So the splicing step need to check the node name.

           Example::

                a -- b -- c --d -- e --f
                     |             |
                     c1 -- d1 -----
                     |             |
                     c2 -- d2 -----
                     |             |
                     c3 -- d3 -----

        For now, the algorithm just support the sub-graph's input /output ops are all in pattern.
        You can set the sub-graph input as (), but the results need you to check. Mostly, this
        sub-graph is a part of the pattern.
        As for pattern match / search, apply dfs to every graph list, then check the sub-graph's
        connection with the main computation flow. The idx would make the returned string list
        with right order.
    """

    def _search_subgraph(subgraph):
        """Parse the pattern_list and match sub-graph."""
        p_subgraph = [c[1] for c in subgraph]
        subgraph_idx = [c[0] for c in subgraph]
        m_subgraph = search_straight_pattern(p_subgraph, graph)
        return (m_subgraph, subgraph_idx)

    def _has_duplicated_names_in_main_chain(sub_chain, main_chain, sub_chain_node_idx, has_head):
        """Avoid splicing error when sub_graphs with totally same op types and tail node."""
        main_chain_node_names = [main_chain[i][0] for i in main_chain.keys()]
        ret_flag = False
        if has_head:
            sub_chain_node_names = sub_chain[1:-2]
            sub_chain_node_index = sub_chain_node_idx[1:-1]
        else:
            sub_chain_node_names = sub_chain[:-2]
            sub_chain_node_index = sub_chain_node_idx[:-1]
        for i in range(len(sub_chain_node_names)):
            index = sub_chain_node_index[i]
            if index in main_chain.keys():
                continue
            else:
                name = sub_chain_node_names[i]
                if name in main_chain_node_names:
                    ret_flag = True
                    break
                else:
                    continue

        return ret_flag

    def _check_subgraph(iter_ret, m_subgraph, sub_graph_idx, has_head):
        """Splicing the main_chain and sub_chain."""
        flag = [0] * len(iter_ret)
        for each_sub in m_subgraph:
            for i in range(len(iter_ret)):
                if flag[i] == 0:
                    # get the sub-graph head and tail ops' names
                    tail_idx = sub_graph_idx[-1]
                    tail_name = iter_ret[i][tail_idx][0]
                    if has_head:
                        head_idx = sub_graph_idx[0]
                        head_name = iter_ret[i][head_idx][0]
                        if (each_sub[0] == head_name) and (each_sub[-2] == tail_name):
                            # check sub-graph name (#7)
                            if _has_duplicated_names_in_main_chain(each_sub, iter_ret[i],
                                                                   sub_graph_idx, has_head):
                                continue
                            for j in range(1, len(each_sub) - 1):
                                iter_ret[i][sub_graph_idx[j]] = [each_sub[j], each_sub[-1][j]]
                                # each sub-graph can be matched more than one main_chain
                            flag[i] = 1
                            # break

                    else:
                        # need to improve the implementation of unknown input node
                        if each_sub[-2] == tail_name:
                            # check sub-graph name (#7)
                            if _has_duplicated_names_in_main_chain(each_sub, iter_ret[i],
                                                                   sub_graph_idx, has_head):
                                continue
                            for j in range(0, len(each_sub) - 1):
                                iter_ret[i][sub_graph_idx[j]] = [each_sub[j], each_sub[-1][j]]
                                # each sub-graph just can be matched more than one main_chain
                            flag[i] = 1
                            # break
                else:
                    continue
        return iter_ret

    def _rm_duplicated_rets(results):
        """Remove the duplicated results due to the complicated symmetric issues.

        Note:
            lists may have same nodes names between each other
            if has symmetric chains, they must appear consecutively
            just keep the first one
        """
        if len(results) <= 1:
            return results
        keep_index = []
        i = 0
        length = len(results)
        while i < length:
            keep_index.append(i)
            ret_a = set(results[i][:-1])
            i += 1
            start = i
            for j in range(start, length):
                ret_b = set(results[j][:-1])
                if ret_a == ret_b:
                    i += 1
                else:
                    break

        final_results = [results[k] for k in keep_index]
        return final_results

    assert len(pattern_list) > 0, "The input patten_list can not be empty!"
    main_chain = pattern_list[0]
    pattern_length = main_chain[-1][0] + 1
    m_main_chain, main_chain_idx = _search_subgraph(main_chain)
    # if has sub-graph, like LayerNorm
    if len(pattern_list) > 1:
        if m_main_chain == []:
            return m_main_chain
        m_result = []
        iter_ret = []
        # print(len(m_main_chain))
        for v in m_main_chain:
            tmp = {}
            # op.type list
            # tmp[-1] = v[-1]
            # op_idx : [op_name, op_type]
            i = 0
            for idx in main_chain_idx:
                tmp[idx] = [v[i], v[-1][i]]
                i += 1
            assert i == len(v) - 1, "error occurs in dict converting"
            iter_ret.append(tmp)
        # is_symmetric = True  # False
        has_one_no_head = False
        for subgraph in pattern_list[1:]:
            has_head = True
            subgraph_ = copy.deepcopy(subgraph)
            if subgraph[0] == ():
                has_head = False
                has_one_no_head = True
                subgraph_.remove(())
            m_subgraph, subgraph_idx = _search_subgraph(subgraph_)
            if len(m_subgraph) == 0:
                return []
            iter_ret = _check_subgraph(iter_ret, m_subgraph, subgraph_idx, has_head)
        if has_one_no_head:
            logger.debug(
            "Does not completely support this pattern: {} now, please check the output results."\
            .format(pattern_list))
        pattern_max_len = 1
        for each_ret in iter_ret:
            pattern_max_len = max(pattern_max_len, len(each_ret))
        # if pattern is symmetric but the graph has not this pattern
        # for example the pattern just has a half chain of U-shape pattern
        # a--b--c
        # a--b--^    but just has a--b--c
        if pattern_max_len == len(main_chain) or pattern_max_len < pattern_length:
            return []
        for each_ret in iter_ret:
            if len(each_ret) < pattern_max_len:
                continue
            tmp = []
            tmp_op = []
            for k in sorted(each_ret):
                # append name
                tmp.append(each_ret[k][0])
                tmp_op.append(each_ret[k][1])
            # append op_type
            tmp.append(tmp_op)
            m_result.append(tmp)

        # de-duplication due to the symmetric chains
        # take the first one out
        m_result = _rm_duplicated_rets(m_result)
        return m_result

    # if has no sub-graph, like MatMul_BiasAdd
    else:
        return m_main_chain


def construct_node(node_name, op_type, input_tensors=[], output_tensors=[], attr=OrderedDict()):
    """Construct node with engine op_type.

    Args:
        node_name: string, name of the node
        op_type: string, type of the node
        input_tensors: list, contains the input tensors of the node
        output_tensors: list, contains the output tensors of the node

    Returns:
        new_node: Operator class
    """
    from .ops.op import OPERATORS, Operator
    from .ops.tensor import Tensor
    if op_type not in OPERATORS.keys():
        new_node = OPERATORS["OpAny"]()
    else:
        new_node = OPERATORS[op_type]()
    new_node.construct(node_name,
                       op_type,
                       input_tensors=input_tensors,
                       output_tensors=output_tensors,
                       attr=attr)
    return new_node


def insert_pattern(target_node_names, new_nodes, graph):
    """Replace the specific pattern matched from the new constructed graph with new pattern.

    Args:
        target_node_names: A string list ccontains the names of nodes that will be replaced
        new_nodes: a list contains nodes with Operator class
        graph: The Graph class

    Returns:
        graph: The Graph class which some nodes inside have been replaced.
    """
    # empty node_names
    if len(target_node_names) == 0:
        return graph

    # only one node
    elif len(target_node_names) == 1:
        node_name = target_node_names[0]
        index = graph.get_node_id(node_name)
        graph.remove_nodes([node_name])

    else:
        # check the order
        # in some conditions, not every name in the target_node_names exists in the graph,
        # the node may be removed last time. For example,
        # a--b--c---d0--e0--f0
        #        ---d1--e1--f1
        #        ---d2--e2--f2
        #        ....
        #        ---dn--en--fn
        # the [dn--en--fn] has same op_type, and the algorithm finds n results in graph,
        # but at the first replace iteration, the [a,b,c] has been removed, so the next
        # iterations missing the [a,b,c]. Need to get the real head and tail node names.
        exist_node_index = []
        for i in range(len(target_node_names)):
            try:
                j = graph.get_node_id(target_node_names[i])
                exist_node_index.append([i, j])
            except BaseException:
                continue
        exist_node_index = sorted(exist_node_index, key=lambda x: x[1])
        exist_node_names = [target_node_names[i[0]] for i in exist_node_index]

        head_name = exist_node_names[0]
        tail_name = exist_node_names[-1]
        head_id = graph.get_node_id(head_name)
        tail_id = graph.get_node_id(tail_name)
        # in the graph.nodes[head_id:tail_id+1], there may be some other nodes
        # have input tensors of new_node
        index = head_id
        i = 0

        while i < len(exist_node_names):
            if exist_node_names[i] == graph.nodes[index].name:
                graph.remove_nodes([exist_node_names[i]])
                i += 1
            else:
                # if not has extra input tensors
                if len(exist_node_names) == (tail_id - head_id + 1):
                    raise ValueError("The target nodes have node {} while graph has node {}."\
                    .format(exist_node_names[i], graph.nodes[index].name))
                # if has extra input tensors
                else:
                    index += 1
    # insert new_nodes
    graph.insert_nodes(index, new_nodes)

    return graph


def pattern_mapping(pattern_name, mapping_dict, graph):
    """The pattern mapping function.

    Args:
        pattern_name:  the name of the customized pattern representation, for example, 'LayerNorm'
        mapping_dict: a element in mapping_config[pattern_name], config for pattern mapping.
        graph: Graph class.

    Returns:
        tuple, the first element is the new nodes insert start idx, the second element is a new
        node list, the third is a list contains required old nodes need to be returned from origin
        pattern.

    Example of mapping_dict::

        {'patterns': {'in': [(0, 'Reshape), ...], 'out':[(0, 'PaddingSequence')]},
         'search_mode': op_type,
         'node_names': {0: 'embeddings/reshape', 1: 0, ...},
         'input_tensors': {0:[{0:[0]}, [[0], 1]], 1:[{1:[0], 2:[1,2]},[[0,1], 3]],
                           2:[{},[[],1]], ..., m:[{'input_data':[1]}, [[0],1]},
         'output_tensors': {2:[{0:[0]}, [[0],1]], ...},
         'returns': [0, 1, 2],
        }                        # one representation of this pattern

    'patterns': give the pattern representations before ('in') and after ('out') fusion. See the
                search_pattern() function for more details about pattern representation.
    'search_mode': 'op_type' or 'node_name'. If set it as op_type, the algorithm will search
                in_pattern in graph. If set node_name, means in_pattern is just representing the
                search result. For example:
                in_pattern is [[(0, 'input_ids'), (1, 'segment_ids'), (2, 'input_mask')]]
                out_pattern is [[(0, 'Input')]]
    'node_names': set node name for each node in pattern after fusion. Key means the node idx,
                the value must be string or int (idx). If the value is the string, just use it as
                the node's name. If the value is the idx, use the name of idx-th node in the
                pattern berfore fusion. If the in_pattern has n match_results in the graph, it
                will add "_n" after the name, for example, the new node name should be
                "embeddings/reshape_0" after mapping of the first match_result.
    'input_tensors': the input_tensors of patterns before or after fusion should be same. The key
                in the dict is the idx of the new node, and the first dict in the value list means
                where this tensor get from the pattern before fusion, and the second means where
                this tensor go to the pattern after fusion. For example, in '0:[{0:[0]},
                [[0], 1]]', '0' in the key means it's the first new node in out_pattern, '{0:[0]}'
                means the tensor is the first tensor of the first node in in_pattern, '[[0], 1]'
                means the first new node's first input_tensor is the tensor and this node has
                total 1 input_tensor. So the first element in the value gives the source info of
                input_tensors, the second gives the dest info of the input_tensors.However,
                sometimes source info has the form like '{1:[0], 2:[1,2]}', the '[1,2]' means the
                idx of tensor is not sure, maybe 1 or 2. It will happens to some sepcial op, like
                'BiasAdd', its 'bias' tensor maybe in unfixed location. If some input_tensors only
                can get from other node outside the pattern, you can just specify it by give the
                node name in graph.
    'output_tensors': the output_tensors of patterns before or after fusion should be same. The
                representtaion is same meaning of 'input_tensors'.
    'returns': set the node idx, and return the idx-th node of pattern before fusion. Sometimes
                need these nodes for writing node attributes in pattern after fusion. If don't
                need return, set the value as [].

    Note that the pattern after fusion (n->n / n->1)is must be sequence pattern like
    [a->b->c->d->e], or [a]. That means if one pattern is too complicated, or the pattern after
    fusion is too complicated, you had better decompose it.

    """

    def _get_pattern_info():
        """Search pattern and get the in_pattern match_result and info for out_pattern."""
        in_pattern = mapping_dict['patterns']['in']
        search_mode = mapping_dict['search_mode']
        assert search_mode in ['op_type', 'node_name'], 'Unsupported mode'
        in_match_result = []
        if search_mode == 'op_type':
            in_match_result = search_pattern(in_pattern, graph)
            num_match = len(in_match_result)

            # WordEmbeddings and TokenTypeEmbeddings is same in some models.
            # Distinguish them by data size
            if num_match == 2 and pattern_name == "WordEmbeddings":
                if len(in_match_result[0]) == 5 and in_match_result[0][-1][1] == "Gather":
                    gather0 = graph.get_node_by_name(in_match_result[0][1])
                    gather1 = graph.get_node_by_name(in_match_result[1][1])
                    if gather0.input_tensors[0].data.size > gather1.input_tensors[0].data.size:
                        in_match_result = [in_match_result[0]]
                    else:
                        in_match_result = [in_match_result[1]]
            # transpose output may be used by another nodes (or as a model output)
            # check its connections and return new matched results
            if num_match > 0 and pattern_name == "TransposeBatchMatMul":
                new_in_match_result = []
                for name_list in in_match_result:
                    keep_flag = True
                    for name in name_list[:-1]:
                        node = graph.get_node_by_name(name)
                        if node.op_type == "Transpose" and (
                                len(node.output_tensors[0].dest_op) > 1 or
                                node.output_tensors[0].name in graph.output_tensors_name):
                            keep_flag = False
                            break
                    if keep_flag:
                        new_in_match_result.append(name_list)
                in_match_result = new_in_match_result
        else:
            # check whether the nodes exit or not
            nodes_exist = True
            in_result = []
            for n in in_pattern[0]:
                n_name = n[1]
                try:
                    _ = graph.get_node_id(n_name)
                    in_result.append(n_name)
                except BaseException:
                    nodes_exist = False
                    break
            if nodes_exist:
                in_result.append([])
                in_match_result.append(in_result)
            else:
                in_match_result = []
        num_match = len(in_match_result)

        new_node_names = []
        name_reference = mapping_dict['node_names']
        input_tensors = []
        input_tensors_reference = mapping_dict['input_tensors']
        output_tensors = []
        output_tensors_reference = mapping_dict['output_tensors']
        ret_old_nodes = []
        returns_inference = mapping_dict['returns']
        for i in range(num_match):
            # get the node_names in out_pattern
            names = []
            for j in range(len(name_reference)):
                tmp = name_reference[j]
                if isinstance(tmp, int):
                    names.append(in_match_result[i][tmp])
                elif isinstance(tmp, str):
                    if num_match == 1:
                        names.append(tmp)
                    else:
                        names.append(tmp + '_' + str(i))
                else:
                    raise ValueError(
                        'Do not support the setted node_names types,it must be int or str,'\
                            'rather than {}.'.format(type(tmp)))
            new_node_names.append(names)

            # get the input_tensors in out_pattern
            in_tensors = []
            for k in range(len(input_tensors_reference)):
                source = input_tensors_reference[k][0]
                dest = input_tensors_reference[k][1]
                tmp = []
                for kv in source:
                    if len(kv) != 0:
                        tmp_k = list(kv.keys())[0]
                        tmp_v = kv[tmp_k]
                        if isinstance(tmp_k, int):
                            node = graph.get_node_by_name(in_match_result[i][tmp_k])
                            assert len(tmp_v) <= 2
                            if len(tmp_v) == 1:
                                v = tmp_v[0]
                            # for support the bias_add, the bias tensor idx may be not always same
                            else:
                                v = tmp_v[0]
                                t0 = node.input_tensors[v]
                                pre_node = graph.get_node_by_name(in_match_result[i][tmp_k - 1])
                                if t0.name == pre_node.output_tensors[0].name:
                                    v = tmp_v[1]
                            tmp.append(copy.deepcopy(node.input_tensors[v]))
                        elif isinstance(tmp_k, str):
                            node = graph.get_node_by_name(tmp_k)
                            tmp.append(copy.deepcopy(node.output_tensors[tmp_v[0]]))
                        else:
                            raise ValueError(
                                'Do not support the setted input_tensors types,'\
                                    'it must be int or str, rather than {}.'.format(type(tmp_k)))

                in_tensors.append([tmp, dest])
            input_tensors.append(in_tensors)

            # get the output_tensors in out_pattern
            out_tensors = []
            for m in range(len(output_tensors_reference)):
                source = output_tensors_reference[m][0]
                dest = output_tensors_reference[m][1]
                tmp = []
                for kv in source:
                    if len(kv) != 0:
                        tmp_k = list(kv.keys())[0]
                        tmp_v = kv[tmp_k]
                        if isinstance(tmp_k, int):
                            node = graph.get_node_by_name(in_match_result[i][tmp_k])
                        else:
                            raise ValueError(
                                'Do not support the setted output_tensors types, it must be int,'\
                                    'rather than {}.'.format(type(tmp_k)))
                        assert len(tmp_v) == 1, 'Output tensor must be specified.'
                        tmp.append(copy.deepcopy(node.output_tensors[tmp_v[0]]))
                out_tensors.append([tmp, dest])
            output_tensors.append(out_tensors)
            # get the returns node in in_pattern for set the attr of node in out_pattern later
            ret_tmp = []
            for idx in returns_inference:
                node = graph.get_node_by_name(in_match_result[i][idx])
                ret_tmp.append(copy.deepcopy(node))
            ret_old_nodes.append(ret_tmp)

        return (in_match_result, new_node_names, input_tensors, output_tensors, ret_old_nodes)

    def _create_out_pattern(new_node_names, input_tensors_list, output_tensors_list):
        """Created the new nodes in out_pattern."""
        from .ops.tensor import Tensor
        out_pattern = mapping_dict['patterns']['out']
        # sequence = mapping_dict['sequence']
        sequence = True if len(out_pattern[0]) > 1 else False
        num_pattern = len(new_node_names)
        new_node_types = []
        new_nodes = []
        for pattern in out_pattern:
            for p in pattern:
                new_node_types.append(p[1])
        # single sequence pattern or several separated nodes
        for i in range(num_pattern):
            one_p = []
            pre_node = None
            for j in range(len(new_node_types)):
                op_type = new_node_types[j]
                node_name = new_node_names[i][j]
                # set the input_tensors
                tensors = input_tensors_list[i][j][0]
                idx_tensors = input_tensors_list[i][j][1][0]
                num_tensors = input_tensors_list[i][j][1][1]
                input_tensors = [0] * num_tensors
                if pre_node is None:
                    assert num_tensors == len(tensors),\
                        'No pre node of {}, please supply the completed input_tensors.'.format(
                            node_name)
                    input_tensors = tensors
                else:
                    assert num_tensors == len(tensors) + 1,\
                        'Only support the sequence pattern for now.'
                    for k in range(len(idx_tensors)):
                        input_tensors[idx_tensors[k]] = tensors[k]
                    input_tensors[input_tensors.index(0)] = copy.deepcopy(
                        pre_node.output_tensors[0])
                # set the output_tensors
                tensors = output_tensors_list[i][j][0]
                idx_tensors = output_tensors_list[i][j][1][0]
                num_tensors = output_tensors_list[i][j][1][1]
                # output_tensors = [0] * num_tensors
                # separated node without connections with other nodes in out_pattern
                if len(tensors) != 0:
                    assert len(tensors) == num_tensors
                    output_tensors = tensors
                # has connections
                else:
                    assert num_tensors == 1, 'Only support op with one output_tensor for now.'
                    output_tensors = [
                        Tensor(name=node_name + ':0',
                               source_op=[node_name],
                               dest_op=[new_node_names[i][j + 1]])
                    ]

                new_node = construct_node(node_name,
                                          op_type,
                                          input_tensors=input_tensors,
                                          output_tensors=output_tensors)
                one_p.append(new_node)

                if sequence:
                    pre_node = new_node

            new_nodes.append(one_p)

        return new_nodes

    def _replace_pattern(in_match_result, new_nodes, graph):
        """Replace the in_pattern with out_pattern."""
        assert len(in_match_result) == len(
            new_nodes), 'out_pattern should have as some num as in_pattern in graph.'
        for i in range(len(in_match_result)):
            each_ret = in_match_result[i][:-1]
            insert_nodes = new_nodes[i]
            graph = insert_pattern(each_ret, insert_nodes, graph)

        return graph

    # 1. check the format of mapping_dict
    mapping_dict = pattern_mapping_conf_validation(mapping_dict)
    # 2. get the necessary info for out_pattern construction
    in_match_result, new_node_names, input_tensors_list, output_tensors_list, \
        ret_old_nodes = _get_pattern_info()
    # 3. create the nodes in out_pattern
    new_nodes = _create_out_pattern(new_node_names, input_tensors_list, output_tensors_list)
    # 4. remove the nodes in in_pattern, and insert the nodes in out_pattern
    graph = _replace_pattern(in_match_result, new_nodes, graph)
    # 5. return graph after pattern mapping
    return (graph, new_node_names, ret_old_nodes)


def list2str(src_perm):
    """Convert the shape list to str for emitting yaml.

    Args:
        src_perm: list, for example [1,2,3,4]

    Returns:
        ret: str, for example '1,2,3,4'
    """
    ret = ','.join(str(i) for i in list(src_perm))
    return ret


def str2list(src_str):
    """Convert the str to shape list.

    Args:
        src_str: for example '1,2,3,4'

    Returns:
        ret: list, for example [1,2,3,4]
    """
    ret = []
    s_list = src_str.split(',')
    ret = [int(i) for i in s_list]
    return ret


def pattern_mapping_conf_validation(conf_dict):
    """The validation of the pattern mapping config."""
    dict_schema = Schema({
    'patterns' : Schema({
        'in' : And(
            list,
            lambda s: all(Schema([(), (int, Or(str, Schema([str])))]).validate(p) for p in s),
            lambda s: all((len(t)==2 for t in p[1:]) for p in s[1:]),
            error='The in pattern must supply the node index and op_type, and only the head node'\
            'in sub-chain can be empty.'
        ),
        'out': Schema([[(int, str)]], error='The out pattern must be straight chain.')
    }),

    'search_mode': Or('op_type', 'node_name', error='Only support op_type or node_name these '\
                        'two modes'),

    'node_names': Schema({
        int: Or(int, str)
    },  error='For node_names, the key is the out node index while the value is the old node ' \
                'index or a specific name'),

    'input_tensors': Schema({
        int: And(
            list,
            lambda s: And(all(list for i in s), len(s)==2, error='The value in input_tensors ' \
            'must be list and has length 2'),
            lambda s: Or(
                And(Schema([{Or(int, str): Schema([int])}]).validate(s[0]),
                    lambda s: Schema([[int], int]).validate(s[1])),

                And(len(s[0])==0, isinstance(s[0], list),
                    lambda s: Schema([[], int]).validate(s[1])),

                error='The first element can be empty list or contains several dict for telling'\
                ' where can get the input_tensor of new_node, while the second element can be'\
                ' empty list or contains several int number for telling the input_tensor index '\
                'of new_node.'
            )
        )
    }),

    'output_tensors': Schema({
        int: And(
            list,
            lambda s: And(all(list for i in s), len(s)==2, error='The value in output_tensors ' \
            'must be list and has length 2'),
            lambda s: Or(
                And(Schema([{int: Schema([int])}]).validate(s[0]),
                    lambda s: Schema([[int], int]).validate(s[1])),

                And(len(s[0])==0, isinstance(s[0], list),
                    lambda s: Schema([[], int]).validate(s[1])),

                error='The first element can be empty list or contains several dict for telling'\
                ' where can get the input_tensor of new_node, while the second element can be'\
                ' empty list or contains several int number for telling the input_tensor index '\
                'of new_node.'
            )
        )
    }),

    'returns': Or(Schema([]), Schema([int]), error='Returns can be empty list or contains some ' \
     'old node index for set attr of new node later'),

    },  ignore_extra_keys=True)

    return dict_schema.validate(conf_dict)


class LazyImport(object):
    """Lazy import python module till use.

    Args:
        module_name (string): The name of module imported later
    """

    def __init__(self, module_name):
        """The module name initialization."""
        self.module_name = module_name
        self.module = None

    def __getattr__(self, name):
        """The __getattr__ function."""
        try:
            self.module = importlib.import_module(self.module_name)
            mod = getattr(self.module, name)
        except:
            spec = importlib.util.find_spec(str(self.module_name + '.' + name))
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)
        return mod

    def __call__(self, *args, **kwargs):
        """The __call__ function."""
        function_name = self.module_name.split('.')[-1]
        module_name = self.module_name.split(f'.{function_name}')[0]
        self.module = importlib.import_module(module_name)
        function = getattr(self.module, function_name)
        return function(*args, **kwargs)


def get_model_fwk_name(model):
    """Detect the input model belongs to which framework.

    Args:
        model (string): framework name that supported by Neural Engine,
                        if there's no available fwk info, then return 'NA'.
    """
    onnx = LazyImport('onnx')
    tf = LazyImport('tensorflow')
    torch = LazyImport('torch')

    def _is_onnxruntime(model):
        """Check if the model is onnxruntime."""
        try:
            if isinstance(model, str):
                graph = onnx.load(model)
                assert (len(graph.graph.node) != 0)
            else:
                graph = model.graph
        except:
            pass
        else:
            return 'onnxruntime'
        return 'NA'

    def _is_tensorflow(model):
        """Check if the model is tensorflow."""
        try:
            if isinstance(model, str):
                graph_def = tf.compat.v1.GraphDef()
                with open(model, 'rb') as f:
                    graph_def.ParseFromString(f.read())
            else:
                graph = model.graph_def
        except:
            pass
        else:
            return 'tensorflow'
        return 'NA'

    def _is_torch(model):
        """Check if the model is torch."""
        try:
            if isinstance(model, str):
                torch.jit.load(model)
        except:
            pass
        else:
            return 'torch'
        return 'NA'

    def _is_neural_engine(model):
        """Check if the model is neural engine."""
        if model and os.path.isdir(model):
            file_list = os.listdir(model)
            is_engine = True
            if len(file_list) == 2:
                for file_name in file_list:
                    file_ext = os.path.splitext(file_name)
                    front, ext = file_ext
                    if ext == ".yaml":
                        is_engine &= True
                    elif ext == ".bin":
                        is_engine &= True
                    else:
                        is_engine &= False
                        logger.error("Please Input yaml and bin for neural engine.")
                        return 'NA'
            else:
                return 'NA'
            if is_engine == True:
                return 'neural engine'
        else:
            return 'NA'

    if isinstance(model, str):
        absmodel = os.path.abspath(os.path.expanduser(model))
        assert os.path.exists(absmodel) or os.path.exists(absmodel+'.pb'), \
            'invalid input path, the file does not exist!'

    checker = [_is_onnxruntime, _is_neural_engine, _is_tensorflow, _is_torch]
    for handler in checker:
        fwk_name = handler(model)
        if fwk_name != 'NA':
            break
    assert fwk_name != 'NA', 'Framework is not detected correctly from model format.'

    return fwk_name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The OutputData Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
import copy


@pattern_registry(pattern_type='OutputData')
class OutputData(Pattern):
    """The OutputData pattern.

    Fuse the original sub-graph into the custom acceleration 'OutputData' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        # make the output_data node in graph
        if model.framework_modeling_config['framework'] == 'torch':
            model_output_tensors = OrderedDict()
            for name in model.output_tensors_name:
                model_output_tensors[name] = None
            for node in model.nodes:
                for output_tensor in node.output_tensors:
                    if output_tensor.name in model.output_tensors_name:
                        model_output_tensors[output_tensor.name] = copy.deepcopy(output_tensor)
            tensors = [v for k, v in model_output_tensors.items() if v]
            output_data_node = util.construct_node('output_data',
                                                'Output',
                                                input_tensors=tensors)
            model.insert_nodes(len(model.nodes), [output_data_node])
            model.nodes[-1].attr = None
            return model

        model_output_tensors = OrderedDict()
        for name in model.output_tensors_name:
            model_output_tensors[name] = None
        for node in model.nodes:
            for output_tensor in node.output_tensors:
                if output_tensor.name in model.output_tensors_name:
                    model_output_tensors[output_tensor.name] = copy.deepcopy(output_tensor)
                else:
                    if not output_tensor.dest_op and node.op_type != 'Input':
                        model_output_tensors[output_tensor.name] = copy.deepcopy(output_tensor)
        tensors = [v for k, v in model_output_tensors.items() if v]
        output_data_node = util.construct_node('output_data',
                                               'Output',
                                               input_tensors=tensors)
        model.insert_nodes(len(model.nodes), [output_data_node])
        model.nodes[-1].attr = None

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Length-Adaptive Transformer
# Copyright (c) 2020-present NAVER Corp.
# Apache License v2.0

"""Evolustion: Provide the evolustionary search for pytorch."""

import csv
import logging
import os
import random
import timeit
import numpy as np
import torch
from intel_extension_for_transformers.optimization.utils.utility import LazyImport
torchprofile = LazyImport("torchprofile")
logger = logging.getLogger(__name__)


def approx_ratio(x, n=12, l=384):
    """Get the approximation ratio."""
    s = 0
    i = l
    for _ in range(n):
        i = int(np.ceil(i * (1 - x)))  # i * x
        s += i
    return s / (n * l)

def inverse(x):
    """Get the inverse number."""
    l, r = 0, 1
    while r - l > 1e-12:
        c = (l + r) / 2
        v = approx_ratio(c)
        l, r = (c, r) if x <= v else (l, c)
    return l


def store2str(gene, macs, score, method, parents=None):
    """Store the parmaters into string."""
    store_str = f"({', '.join(f'{x:3d}' for x in gene)}):"
    store_str += f" {macs} MACs/latency"
    store_str += f" | score {score}"
    store_str += f" | method {method}"
    if parents is not None:
        store_str += f"| parent(s) {parents}"
    return store_str

class Evolution(object):
    """Class of Evolution supports for evolutionary searching."""
    def __init__(
        self,
        model,
        max_seq_length,
        device,
        evaluate,
        lower_constraint=0,
        upper_constraint=None,
        eval_metric='eval_f1'
    ):
        """Init an Evolution instance."""
        self.model = model
        self.max_seq_length = max_seq_length
        self.device = device
        self.evaluate = evaluate

        size = (1, self.max_seq_length)
        self.dummy_inputs = (
            torch.ones(size, dtype=torch.long).to(self.device),
            torch.ones(size, dtype=torch.long).to(self.device),
            torch.zeros(size, dtype=torch.long).to(self.device),
        )
        if self.model.config.model_type == "distilbert":
            self.dummy_inputs = self.dummy_inputs[:2]

        self.lower_constraint = lower_constraint
        self.upper_constraint = upper_constraint

        self.store = {}  # gene: (macs, score, method, parent(s))
        self.population = []
        self.eval_metric=eval_metric

    def load_store(self, store_file):
        """Load from a store file."""
        if not os.path.isfile(store_file):
            return
        with open(store_file, 'r') as f:
            for row in csv.reader(f, delimiter='\t'):
                row = tuple(eval(x) for x in row[:3])
                self.store[row[0]] = row[1:3] + (0, None)

    def save_store(self, store_file):
        """Save into a store file."""
        store_keys = sorted(self.store.keys(), key=lambda x: self.store[x][0])
        with open(store_file, 'w') as f:
            writer = csv.writer(f, delimiter='\t')
            for gene in store_keys:
                writer.writerow([str(gene)] + [str(x) for x in self.store[gene]])

    def get_store(self):
        """Get store."""
        return self.store

    def set_lower_constraint(self, constraint):
        """Setter of lower constraint."""
        self.lower_constraint = constraint

    def set_upper_constraint(self, constraint):
        """Setter of upper constraint."""
        self.upper_constraint = constraint

    def save_population(self, population_file, population):
        """Save population into a file."""
        with open(population_file, 'w') as f:
            writer = csv.writer(f, delimiter='\t')
            for gene in population:
                writer.writerow([str(gene)] + [str(x) for x in self.store[gene]])

    def _ccw(self, gene0, gene1, gene2):
        """Counterclockwise."""
        x0, y0 = self.store[gene0][:2]
        x1, y1 = self.store[gene1][:2]
        x2, y2 = self.store[gene2][:2]
        return (x0 * y1 + x1 * y2 + x2 * y0) - (x0 * y2 + x1 * y0 + x2 * y1)

    def convex_hull(self):
        """The function to calculate convex_hull."""
        hull = self.population[:2]
        for gene in self.population[2:]:
            if self.store[hull[-1]][1] >= self.store[gene][1]:
                continue
            while len(hull) >= 2 and self._ccw(hull[-2], hull[-1], gene) >= 0:
                del hull[-1]
            hull.append(gene)
        return hull

    def pareto_frontier(self):
        """The function to calculate population and are."""
        self.population = sorted(self.population, key=lambda x: self.store[x][:2])

        frontier = [self.population[0]]
        for gene in self.population[1:-1]:
            if self.store[gene][1] > self.store[frontier[-1]][1]:
                if self.store[gene][0] == frontier[-1][0]:
                    del frontier[-1]
                frontier.append(gene)
        frontier.append(self.population[-1])
        self.population = frontier

        area = 0
        for gene0, gene1 in zip(self.population[:-1], self.population[1:]):
            x0, y0 = self.store[gene0][:2]
            x1, y1 = self.store[gene1][:2]
            area += (x1 - x0) * y0
        area /= (self.upper_constraint - self.lower_constraint)
        return self.population, area

    def add_gene(self, gene, macs=None, score=None, method=0, parents=None):
        """Add gene to evolution."""
        if gene not in self.store:
            self.model.eval()
            if self.model.config.model_type == "distilbert":
                bert = self.model.distilbert
            elif self.model.config.model_type == "roberta":
                bert = self.model.roberta
            else:
                assert hasattr(self.model, "bert")
                bert = self.model.bert
            bert.set_length_config(gene)
            bert.set_output_attentions(True)
            macs = macs or torchprofile.profile_macs(self.model, args=self.dummy_inputs)
            if macs < self.lower_constraint:
                return False
            start_time = timeit.default_timer()
            eval_result = self.evaluate()
            evalTime = timeit.default_timer() - start_time
            score = score or eval_result[self.eval_metric]
            self.store[gene] = (macs, score, method, parents)
            logger.info(store2str(gene, macs, score, method, parents))

        macs = self.store[gene][0]
        if macs >= self.lower_constraint \
                and (self.upper_constraint is None or macs <= self.upper_constraint) \
                and gene not in self.population:
            self.population.append(gene)
            return True
        return False

    def mutate(self, mutation_prob, ray=False):
        """Do the mutate."""
        gene = random.choice(self.population)
        mutated_gene = ()
        for i in range(self.model.config.num_hidden_layers):
            if np.random.uniform() < mutation_prob:
                prev = (self.max_seq_length if i == 0 else mutated_gene[i - 1])
                next = (2 if i == self.model.config.num_hidden_layers - 1 else gene[i + 1])
                mutated_gene += (random.randrange(next, prev + 1),)
            else:
                mutated_gene += (gene[i],)
        return self.add_gene(mutated_gene, method=1, parents=(gene,)) if not ray else mutated_gene, (gene,)

    def crossover(self, ray=False):
        """Do the crossover."""
        gene0, gene1 = random.sample(self.population, 2)
        crossovered_gene = tuple((g0 + g1 + 1) // 2 for g0, g1 in zip(gene0, gene1))
        return self.add_gene(crossovered_gene, method=2, parents=(gene0, gene1)) if not ray else \
                 crossovered_gene, (gene0, gene1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .config import (
    AutoDistillationConfig,
    DistillationConfig,
    FlashDistillationConfig,
    TFDistillationConfig,
    NASConfig,
    Provider,
    PruningConfig,
    QuantizationConfig,
    WEIGHTS_NAME,
    DynamicLengthConfig,
    BenchmarkConfig,
    PrunerV2,
    WeightPruningConfig
)
from .distillation import (
    DistillationCriterionMode,
    SUPPORTED_DISTILLATION_CRITERION_MODE,
)
from .mixture.auto_distillation import AutoDistillation
from .model import OptimizedModel
from .nas import NAS
from .optimizer import NoTrainerOptimizer, Orchestrate_optimizer
from .optimizer_tf import TFOptimization
from .pruning import PrunerConfig, PruningMode, SUPPORTED_PRUNING_MODE
from .quantization import QuantizationMode, SUPPORTED_QUANT_MODE
from .utils import metrics
from .utils import objectives
from .utils.utility import LazyImport


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TFOptimization: provides the optimization class for Tensorflow."""
import logging
import pstats
import numpy as np
import os
import time
from neural_compressor import __version__
from neural_compressor.experimental import common
try:
    from neural_compressor.model.model import saved_model_session, get_model_type
except ImportError:
    from neural_compressor.model.tensorflow_model import saved_model_session, get_model_type
from intel_extension_for_transformers.optimization import (DistillationConfig,
                                                           QuantizationConfig,
                                                           PruningConfig,
                                                           AutoDistillation)
from intel_extension_for_transformers.optimization.quantization import QuantizationMode
from intel_extension_for_transformers.optimization.utils.metrics import Metric
from intel_extension_for_transformers.optimization.utils.utility import LazyImport
from packaging import version
from transformers import PreTrainedModel
from transformers.training_args_tf import TFTrainingArguments
from typing import Callable, Optional, List
from .utils.utility_tf import TFDataloader, TMPPATH, TEACHERPATH, get_filepath

from functools import partial

tf = LazyImport("tensorflow")
logger = logging.getLogger(__name__)
logger.setLevel('INFO')


class TFOptimization:
    """TFOptimization is the entry class for Tensorflow to use the optimization techniques in neural compressor."""
    def __init__(self,
                 model: PreTrainedModel,
                 args: TFTrainingArguments,
                 train_dataset=None,
                 eval_dataset=None,
                 compute_metrics: Optional[Callable] = None,
                 criterion=None,
                 optimizer=None,
                 task_type=None,
                 task_id=None,
                 strategy=None):
        """Init a TFOptimziation object.

        Args:
            model: FP32 model specified for low precision tuning
            args: Training arguments for TF
            train_dataset: Training datas
            eval_dataset: Evaluation data
            compute_metrics: Metrics computing function during the evaluation process
            criterion: Tuning criterion
            optimizer: The optimizer you used
            task_type: Used for distributed multi-node settings. Default set as "worker"
            task_id: Used for distributed multi-node settings.
                Set as 0 on the leader node and 1, 2, 3... on the followers
            strategy: Equals to MultiWorkerMirroredStrategy if use distributed distributed multi-node settings,
                otherwise None
        """
        self.model = model
        self.teacher_model = None
        self.component = None
        self.eval_dataset = eval_dataset
        self.train_dataset = train_dataset
        self._eval_func = None
        self._train_func = None
        self.quant_config = None
        self.pruning_config = None
        self.distillation_config = None
        self.pruner = None
        self.quantizer = None
        self.distiller = None
        self.in_training = False
        self._input_names = None
        self._output_names = None
        self._inputs = None
        self.compute_metrics = compute_metrics
        self.args = args
        self.optimizer = optimizer
        self.task_type = task_type
        self.task_id = task_id
        self.criterion = criterion if criterion is not None else \
            self.model.loss if hasattr(self.model, "loss") else None
        self.model.save_pretrained(get_filepath(TMPPATH, self.task_type, self.task_id), saved_model=True)
        _, self.input_names, self.output_names = saved_model_session(
            os.path.join(get_filepath(TMPPATH, self.task_type, self.task_id), "saved_model/1"), input_tensor_names=[],
             output_tensor_names=[])
        self.eval_distributed = False
        self.strategy = strategy

    @property
    def inputs(self):
        """Get the inputs."""
        return self._inputs

    @inputs.setter
    def inputs(self, inputs: dict):
        """Set the inputs."""
        self._inputs = inputs

    @property
    def input_names(self):
        """Get the input names."""
        return self._input_names

    @input_names.setter
    def input_names(self, input_names: List):
        """Set the input names.
        
        Args:
            input_names: the names of inputs.
        """
        self._input_names = input_names

    @property
    def output_names(self):
        """Get the output names."""
        return self._output_names

    @output_names.setter
    def output_names(self, output_names: List):
        """Set the output names.
        
        Args:
            output_names: the names of outputs.
        """
        self._output_names = output_names

    @property
    def eval_func(self):
        """Get the evaluation function."""
        return self._eval_func

    @eval_func.setter
    def eval_func(self, func: Callable):
        """Set the evaluation function.
        
        Args:
            func: evaluation function.
        """
        self._eval_func = func

    @property
    def train_func(self):
        """Get the training function."""
        return self._train_func

    @train_func.setter
    def train_func(self, func: Callable):
        """Set the training function.
        
        Args:
            func: train function.
        """
        self._train_func = func

    @property
    def train_dataset(self):
        """Get the training dataset."""
        return self._train_dataset

    @train_dataset.setter
    def train_dataset(self, train_dataset):
        """Set the training dataset.
        
        Args:
            train_dataset: train dataset.
        """
        assert isinstance(train_dataset, tf.data.Dataset) or train_dataset is None, \
            "train_dataset should be obj of tf.data.Dataset"
        self._train_dataset = train_dataset

    @property
    def eval_dataset(self):
        """Get the evaluation dataset."""
        return self._eval_dataset

    @eval_dataset.setter
    def eval_dataset(self, eval_dataset):
        """Set the evaluation dataset.
        
        Args:
            eval_dataset: evaluation dataset.
        """
        assert isinstance(eval_dataset, tf.data.Dataset) or eval_dataset is None, \
            "eval_dataset should be obj of tf.data.Dataset"
        self._eval_dataset = eval_dataset

    def builtin_eval_func(self, model):
        """Customize Evaluate function to inference the model for specified metric on the validation dataset.

        Args:
            model ([tf.saved_model.load]): The model will be the class of tf.saved_model.load(quantized_model_path).

        Returns:
            [float]: evaluation result, the larger is better.
        """
        model_type = None
        label_ids: np.ndarray = None
        try:
            model_type = get_model_type(model)
        except ValueError:
            logger.info("use keras savedModel")

        num_examples = sum(1 for _ in (
            self._eval_dataset.unbatch() if hasattr(self._eval_dataset, "unbatch") else self._eval_dataset))
        logger.info(f"***** Running Evaluation *****")
        logger.info(f"  Num examples in dataset = {num_examples}")
        logger.info(f"  Batch size = {self.args.per_device_eval_batch_size}")

        if model_type is None:
            preds: np.ndarray = None
            infer = model.signatures["serving_default"]

            for idx, (inputs, labels) in enumerate(self._eval_dataset):
                for name in inputs:
                    inputs[name] = tf.constant(inputs[name].numpy(), dtype=infer.inputs[0].dtype)

                results = infer(**inputs)
                for val in results:
                    if preds is None:
                        preds = results[val].numpy()
                    else:
                        preds = np.append(preds, results[val].numpy(), axis=0)

                if label_ids is None:
                    label_ids = labels[0].numpy() if isinstance(
                        labels, list) else labels.numpy()
                else:
                    label_ids = np.append(
                        label_ids,
                        labels[0].numpy()
                        if isinstance(labels, list) else labels.numpy(),
                        axis=0)
            test_predictions = {"logits": preds}
            eval_metrics = self.compute_metrics(test_predictions, label_ids)
            acc = eval_metrics["accuracy"]
            return acc
        else:  # pragma: no cover
            from neural_compressor.adaptor.tf_utils.util import get_tensor_by_name
            input_tensor = [get_tensor_by_name(\
                model, x) for x in self.input_names]
            output_tensor = [get_tensor_by_name(\
                model, x) for x in self.output_names]

            logger.info("Start to evaluate the TensorFlow model.")

            total_time = 0
            config = tf.compat.v1.ConfigProto()
            config.use_per_session_threads = 1
            config.inter_op_parallelism_threads = 1
            sess = tf.compat.v1.Session(graph=model, config=config)
            feed_dict = {}
            label_ids: np.ndarray = None
            preds: np.ndarray = None
            for idx, (inputs, labels) in enumerate(self._eval_dataset):
                assert len(input_tensor) == len(inputs), \
                    'inputs len must equal with input_tensor'
                feed_dict = {}
                for name in inputs:
                    for tensor in input_tensor:
                        pos = tensor.name.rfind(":")
                        t_name = tensor.name if pos < 0 else tensor.name[:pos]
                        if name == t_name:
                            feed_dict[tensor] = inputs[name].numpy()
                            break

                start = time.time()
                logits = sess.run(output_tensor, feed_dict)
                total_time += time.time() - start
                if not self.args.prediction_loss_only:
                    if isinstance(logits, tuple):
                        logits = logits[0]

                    if isinstance(labels, tuple):
                        labels = labels[0].numpy()

                    if isinstance(logits,
                                list) and len(logits) > 1:  # pragma: no cover
                        for val in logits:
                            if preds is None:
                                preds = val
                            else:
                                preds = np.append(preds, val, axis=0)

                        for val in labels:
                            if label_ids is None:
                                label_ids = val.numpy()
                            else:
                                label_ids = np.append(label_ids,
                                                    val.numpy(),
                                                    axis=0)
                    else:
                        if preds is None:
                            preds = logits[0] if isinstance(logits,
                                                            list) else logits
                        else:
                            preds = np.append(
                                preds,
                                logits[0] if isinstance(logits, list) else logits,
                                axis=0)

                        if label_ids is None:
                            label_ids = labels[0].numpy() if isinstance(
                                labels, list) else labels.numpy()
                        else:
                            label_ids = np.append(
                                label_ids,
                                labels[0].numpy()
                                if isinstance(labels, list) else labels.numpy(),
                                axis=0)

            if self.compute_metrics is not None and preds is not None and label_ids is not None:
                try:
                    loss = self.criterion(
                        label_ids, preds) if self.criterion is not None else None
                except Exception as e:  # pragma: no cover
                    logger.info(e)
                    logger.info("There is no loss function or loss compute error, \
                                    Please compute loss in compute_metrics function"
                                )
                    loss = None
                results = self.compute_metrics({"logits": preds}, label_ids)
                if loss is not None:
                    results["loss"] = loss.numpy()

                if isinstance(self.metrics, list):
                    nums = len(self.metrics)
                    for metric in self.metrics:
                        assert metric.name in results.keys(), \
                            "Please set metric from {}".format(results.keys())
                    if nums == 1:
                        result = results.get(self.metrics[0].name)
                    else:  # pragma: no cover
                        result = 0
                        for metric in self.metrics:
                            assert metric.weight_ratio is not None, \
                                "Please set weights for metric if you want to use more than one metric"
                            result += results[metric.name] * metric.weighted
                    logger.info("metric Accuracy: {}".format(result))
                elif isinstance(self.metrics, Metric):
                    assert self.metrics.name in results.keys(), \
                            "Please set metric from {}".format(results.keys())
                    result = results.get(self.metrics.name)
                    logger.info("metric Accuracy: {}".format(result))
                else:  # pragma: no cover
                    assert False, "Please set the correct metrics format from the README"
            else:
                result = 0
            logger.info("Throughput: {} samples/sec".format(num_examples / total_time))
            return result

    def init_quantizer(
        self,
        quant_config,
    ):
        """Init a Quantization object with config.
        
        Args:
            quant_config: quantization config.
        """
        from neural_compressor.experimental import Quantization

        self.quant_config = QuantizationConfig() if quant_config is None else quant_config
        self.quant_config.framework = "tensorflow"
        self.metrics = self.quant_config.metrics

        quantizer = Quantization(self.quant_config.inc_config)
        quantizer.model = common.Model(
            os.path.join(get_filepath(TMPPATH, self.task_type, self.task_id),"saved_model/1"), modelType="saved_model")

        self.quantizer = quantizer
        return quantizer

    def _inc_quantize(
        self,
        quant_config,
    ):
        """Do the quantization.
        
        Args:
            quant_config: quantization config.
        """
        if self.quantizer is None:
            self.init_quantizer(quant_config=quant_config)
        if self._eval_func is not None:
            self.quantizer.eval_func = self._eval_func
        else:
            assert self.metrics is not None, \
                "Please pass the metrics to QuantizationConfig.metrics!"
            self.quantizer.eval_func = self.builtin_eval_func

        if self.quant_config.approach == QuantizationMode.POSTTRAININGSTATIC.value:
            if self._train_dataset is not None:
                self.quantizer.calib_dataloader = TFDataloader(
                    self._train_dataset,
                    batch_size=self.args.per_device_train_batch_size)
            elif self._eval_dataset is not None:
                self.quantizer.calib_dataloader = TFDataloader(
                    self._eval_dataset,
                    batch_size=self.args.per_device_eval_batch_size)
            else:  # pragma: no cover
                assert False, "Please pass calibration dataset to TFNoTrainerOptimizer.calib_dataloader"
        elif self.quant_config.approach == QuantizationMode.QUANTIZATIONAWARETRAINING.value:   # pragma: no cover
            assert False, \
                "Unsupport quantization aware training for tensorflow framework"

        opt_model = self.quantizer.fit()
        opt_model.save(self.args.output_dir)
        logger.info(
            "quantized model have saved to {}".format(self.args.output_dir)
        )
        return opt_model.model

    def quantize(
        self,
        quant_config: QuantizationConfig = None,
        eval_func: Optional[Callable] = None,
        train_func: Optional[Callable] = None,
        train_dataset=None,
        eval_dataset=None,
    ):
        """Prepare for invoking INC quantize function.
        
        Args:
            quant_config: quantization config.
            eval_func: evaluation function.
            train_func: train function.
            train_dataset: train dataset.
            eval_dataset: evaluation dataset.
        """
        if eval_func is not None:
            self._eval_func = eval_func
        if train_func is not None:
            self._train_func = train_func
        if train_dataset is not None:
            self.train_dataset = train_dataset

        if eval_dataset is not None:
            self.eval_dataset = eval_dataset

        return self._inc_quantize(quant_config=quant_config)

    def init_pruner(
        self,
        pruning_config=None,
    ):
        """Init a Pruning object with config.
        
        Args:
            pruning_config: pruning config.
        """
        from neural_compressor.experimental import Pruning
        if pruning_config.framework != 'tensorflow':
            logger.warning('pruning_config.framework is {}, should be tensorflow'.format(pruning_config.framework))
            pruning_config.framework = 'tensorflow'
        self.pruning_config = pruning_config
        self.metrics = self.pruning_config.metrics

        assert isinstance(self.pruning_config, PruningConfig), \
            "please pass a instance of PruningConfig to trainer.prune!"

        pruner = Pruning(self.pruning_config.inc_config)
        pruner.model = os.path.join(get_filepath(TMPPATH, self.task_type, self.task_id), "saved_model/1")
        pruner.model.model_type = "saved_model"

        self.pruner = pruner
        self.component = pruner
        return pruner

    def prune(
        self,
        pruning_config=None,
        eval_func: Optional[Callable] = None,
        train_func: Optional[Callable] = None,
        train_dataset=None,
        eval_dataset=None,
    ):
        """Do the pruning.
        
        Args:
            pruning_config: pruning config.
            eval_func: evaluation function.
            train_func: train function.
            train_dataset: train dataset.
            eval_dataset: evaluation dataset.
        """
        if self.pruner is None:
            self.init_pruner(pruning_config=pruning_config)
        if eval_func is not None:
            self.eval_func = eval_func
        if train_func is not None:
            self.train_func = train_func

        if train_dataset is not None:
            self.train_dataset = train_dataset

        if eval_dataset is not None:
            self.eval_dataset = eval_dataset

        if self._eval_func is not None:
            self.pruner.eval_func = self._eval_func
        else:
            assert self.metrics is not None, \
                "Please pass the metrics to PruningConfig.metrics!"
            self.pruner.eval_func = self.builtin_eval_func

        if self.train_func is not None:
            if version.parse(__version__) <= version.parse("1.12"):
                self.pruner.pruning_func = self._train_func
            else:
                self.pruner.train_func = self._train_func
        else:
            if version.parse(__version__) <= version.parse("1.12"):
                self.pruner.pruning_func = self.build_train_func
            else:
                self.pruner.train_func = self.build_train_func

        opt_model = self.pruner.fit()
        stats, sparsity = opt_model.report_sparsity()
        logger.info(stats)
        logger.info(sparsity)

        opt_model.save(self.args.output_dir)
        logger.info(
            "pruned model have saved to {}".format(self.args.output_dir)
        )
        return opt_model.model

    def init_distiller(
        self,
        distillation_config,
        teacher_model: PreTrainedModel,
    ):
        """Init a Distillation object with config and the teacher model.
        
        Args:
            distillation_config: distillation config.
            teacher_model: set the teacher model.
        """
        from neural_compressor.experimental import Distillation
        assert isinstance(distillation_config, DistillationConfig), \
            "please pass a instance of DistillationConfig to trainer.distill!"

        def train_step(data):
            if len(data) == 3:
                x, y, sample_weight = data  # pragma: no cover
            else:
                sample_weight = None
                x, y = data
            with tf.GradientTape() as tape:
                y_pred = self.model(x)
                teacher_outputs = self.distiller.criterion.teacher_model_forward(
                    input=x, teacher_model=teacher_model)

                loss = self.model.compute_loss(x, y, y_pred, sample_weight)
                # _on_after_compute_loss(self, input, student_output, student_loss, teacher_output=None)
                # TODO: check, combile
                loss = self.distiller.on_after_compute_loss(
                    x, y_pred.logits, loss, teacher_outputs.logits)
            self.model._validate_target_and_loss(y, loss)
            # Run backwards pass.
            self.model.optimizer.minimize(loss,
                                          self.model.trainable_variables,
                                          tape=tape)
            return self.model.compute_metrics(x, y, y_pred, sample_weight)

        self.model.train_step = train_step
        # re-compile
        self.model.compile(
            optimizer=self.model.optimizer,
            loss=self.model.loss,
            metrics=self.model.compiled_metrics._user_metrics
            )

        if distillation_config.framework != 'tensorflow':
            logger.warning(
                'distillation_config.framework is {}, should be tensorflow'.
                format(distillation_config.framework))
            distillation_config.framework = 'tensorflow'
        self.distillation_config = distillation_config
        self.metrics = self.distillation_config.metrics
        self.teacher_model = teacher_model

        distiller = Distillation(self.distillation_config.inc_config)
        distiller.model = os.path.join(TMPPATH, "saved_model/1")
        distiller.model.model_type = "saved_model"
        self.teacher_model.save_pretrained(TEACHERPATH, saved_model=True)
        distiller.teacher_model = os.path.join(TEACHERPATH, "saved_model/1")
        distiller.teacher_model.model_type = "saved_model"

        self.distiller = distiller
        self.component = distiller
        return distiller

    def distill(
        self,
        distillation_config,
        teacher_model: PreTrainedModel,
        eval_func: Optional[Callable] = None,
        train_func: Optional[Callable] = None,
    ):
        """Do the distillation.
        
        Args:
            distillation_config: distillation config.
            teacher_model: set the teacher model.
            eval_func: evaluation function.
            train_func: train function.
        """
        if self.distiller is None:
            self.init_distiller(
                distillation_config=distillation_config,
                teacher_model=teacher_model,
            )
        if eval_func is not None:
            self._eval_func = eval_func
        if train_func is not None:
            self._train_func = train_func
        else:
            self._train_func = self.build_train_func

        self.distiller.eval_func = self._eval_func
        self.distiller.train_func = self._train_func
        self.distiller.create_criterion()

        opt_model = self.distiller.fit()
        opt_model.save(self.args.output_dir)
        logger.info(
            "distilled model have saved to {}".format(self.args.output_dir)
        )

        return opt_model.model

    def model_builder_builtin(self, arch_paras=None, model_cls=None):
        """Specify model_cls to use the built-in model builder.
        
        Args:
            arch_paras: architecture parameters.
            model_cls: model information.
        """
        config = self.model.config
        if arch_paras is not None:
            assert isinstance(arch_paras, dict), "Expect arch_paras to be a dict."
            for k in arch_paras:
                if hasattr(config, k):
                    config.__setattr__(k, arch_paras[k])
                    # for MobileBERT, 'intra_bottleneck_size' is associated with
                    # 'true_hidden_size', and must have the same values.
                    if k == 'intra_bottleneck_size':
                        config.__setattr__('true_hidden_size', arch_paras[k])
        return model_cls.from_config(config)


    def autodistill(
        self,
        autodistillation_config,
        teacher_model: PreTrainedModel,
        model_builder: Optional[Callable] = None,
        model_cls: Optional[Callable] = None,
        eval_func: Optional[Callable] = None,
        train_func: Optional[Callable] = None
        ):
        """Do the auto distillation.
        
        Args:
            autodistillation_config: autodistillation config.
            teacher_model: set the teacher model.
            model_builder: the configuration of build in model.
            model_cls: the model information.
            eval_func: evaluation function.
            train_func: train function.
        """
        self.autodistillation_config = autodistillation_config
        if model_builder is None:
            assert model_cls is not None, "Must specify model_cls to use the built-in " + \
                "model_builder, e.g. model_cls=AutoModelForPreTraining, or you can use " + \
                "the customized model_builder."
            model_builder = partial(self.model_builder_builtin, model_cls=model_cls)
        agent = AutoDistillation(model_builder, self.autodistillation_config, framework='tensorflow')

        def train_func_builtin(model):
            """Get the build in train function.

            Args:
                model (object): the input model
            """
            def run_distillers(
                model,
                distillers,
                train_steps,
                block_names,
                presentation='flash distillation'
            ):
                """Get the distiller.

                Args:
                    model (object): the input model.
                    distillers: distillers.
                    train_steps: number of train steps.
                    block_names: the name of the block.
                    presentation: presentation format.
                """
                for i, elements in enumerate(zip(distillers, train_steps, block_names)):
                    distiller, ts, bln = elements
                    logger.info(' '.join(
                        ['=' * 30, 'Step {} of'.format(i + 1), presentation, '=' * 30]))

                    def train_step(data):
                        if len(data) == 3:
                            x, y, sample_weight = data  # pragma: no cover
                        else:
                            sample_weight = None
                            x, y = data
                        with tf.GradientTape() as tape:
                            y_pred = model(x)
                            teacher_outputs = distiller.criterion.teacher_model_forward(
                                input=x, teacher_model=teacher_model)

                            loss = model.compute_loss(x, y, y_pred, sample_weight)
                            # _on_after_compute_loss(self, input, student_output, student_loss, teacher_output=None)
                            # TODO: check, combile
                            loss = distiller.on_after_compute_loss(
                                x, y_pred.logits, loss, teacher_outputs.logits)
                        model._validate_target_and_loss(y, loss)
                        # Run backwards pass.
                        optimizer = self.model.optimizer
                        optimizer.minimize(
                            loss,
                            model.trainable_variables,
                            tape=tape)
                        return model.compute_metrics(x, y, y_pred, sample_weight)

                    model.save_pretrained(get_filepath(TMPPATH, self.task_type, self.task_id), saved_model=True)

                    # re-build optimizer
                    opt_kwargs = {}
                    for k, v in self.model.optimizer.__dict__.items():
                        if not k.startswith('_'):
                            opt_kwargs[k] = v
                    optimizer = self.model.optimizer.__class__(**opt_kwargs)
                    if self.strategy:  # pragma: no cover
                        with self.strategy.scope():
                            model = model_cls.from_pretrained(get_filepath(TMPPATH, self.task_type, self.task_id))
                            model.compile(
                                    optimizer=optimizer,
                                    loss=self.model.loss,
                                    metrics=self.model.compiled_metrics._user_metrics
                                    )
                            model.train_step = train_step
                    else:
                        model.train_step = train_step
                        model.compile(
                            optimizer=optimizer,
                            loss=self.model.loss,
                            metrics=self.model.compiled_metrics._user_metrics)
                    self.model = model

                    distiller.model = os.path.join(TMPPATH, "saved_model/1")
                    distiller.model.model_type = "saved_model"
                    teacher_model.save_pretrained(TEACHERPATH, saved_model=True)
                    distiller.teacher_model = os.path.join(TEACHERPATH, "saved_model/1")
                    distiller.teacher_model.model_type = "saved_model"

                    if eval_func is not None:
                        self._eval_func = eval_func
                    else:
                        self._eval_func = self.builtin_eval_func
                    if train_func is not None:
                        self._train_func = train_func
                    else:
                        self._train_func = self.build_train_func

                    distiller.eval_func = self._eval_func
                    distiller.train_func = self._train_func
                    distiller.create_criterion()

                    self.component = self.distiller = distiller

                    opt_model = distiller.fit()
                    opt_model.save(self.args.output_dir)
                    return opt_model

            agent.create_distillers()
            # run flash_distillers
            ori_model = model
            if agent.flash_distillers:
                model = run_distillers(ori_model, agent.flash_distillers,
                                       agent.flash_train_steps,
                                       agent.flash_block_names)
            # run regular_distillers
            if agent.regular_distillers:
                model = run_distillers(ori_model,
                                       agent.regular_distillers,
                                       agent.regular_train_steps,
                                       agent.regular_block_names,
                                       presentation='regular distillation')
            return model.model

        def eval_func_builtin(model):
            """Get the build in evaluation function.

            Args:
                model (object): the input model
            """
            if self._eval_func:
                result = self._eval_func(model)
            else:
                result = self.builtin_eval_func(model)  # pragma: no cover
            return {'metric': result}

        agent.framework = 'tensorflow'
        agent.train_func = train_func \
            if train_func else train_func_builtin
        agent.eval_func = eval_func \
            if eval_func else eval_func_builtin
        # pylint: disable=E1101
        os.makedirs(self.args.output_dir, exist_ok=True)
        return agent.search(self.args.output_dir, model_cls)

    def build_train_func(self, model):
        """Build the training function for pruning or distillation.
        
        Args:
            model (object): the input model
        """
        tf.random.set_seed(1)
        epochs = 1

        component = self.component
        prune_model = self.model
        model_path = get_filepath(TMPPATH, self.task_type, self.task_id)

        if 'distillation' in self.component.cfg:
            epochs = max(epochs, self.component.cfg.distillation.train.get("epoch", 1))
            hooks = self.component.hooks
        if 'pruning' in self.component.cfg:
            epochs = max(epochs, self.component.cfg.pruning.train.get("epoch", 1))
            callbacks = self.pruner.callbacks
            hooks = callbacks['tf_pruning'](self.pruner.model, self.model,
                                            self.pruner.hooks)

        class callback(tf.keras.callbacks.Callback):
            def on_train_begin(self, logs=None):
                if version.parse(__version__) <= version.parse("1.12"):
                    hooks['pre_epoch_begin']()  # pragma: no cover
                else:
                    hooks['on_train_begin']()

            def on_train_end(self, logs=None):
                if version.parse(__version__) <= version.parse("1.12"):
                    hooks['post_epoch_end']()  # pragma: no cover
                else:
                    hooks['on_train_end']()

            def on_epoch_begin(self, epoch, logs=None):
                # pylint: disable=E1121
                hooks['on_epoch_begin'](epoch)

            def on_epoch_end(self, epoch, logs=None):
                component.model._session = None
                prune_model.save_pretrained(model_path, saved_model=True)
                component.model = os.path.join(model_path, "saved_model/1")
                component.model.model_type = "saved_model"
                component.model.sess
                hooks['on_epoch_end']()

            # pylint: disable=E1121
            def on_train_batch_begin(self, batch, logs=None):
                if version.parse(__version__) <= version.parse("1.12"):
                    hooks['on_batch_begin'](batch)  # pragma: no cover
                else:
                    hooks['on_step_begin'](batch)

            def on_train_batch_end(self, batch, logs=None):
                if version.parse(__version__) <= version.parse("1.12"):
                    hooks['on_batch_end']()  # pragma: no cover
                else:
                    hooks['on_step_end']()

        self.model.fit(self.train_dataset,
                       validation_data=self.eval_dataset,
                       epochs=epochs,
                       callbacks=[callback()])
        self.component.model._session = None
        self.model.save_pretrained(get_filepath(TMPPATH, self.task_type, self.task_id), saved_model=True)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
from .. import graph_utils as util
from .. import logger


@pattern_registry(pattern_type='GroupNorm')
class GroupNorm(Pattern):
    """
    The input channels are separated into num_groups groups, each containing num_channels / 
    num_groups channels. Each group is calculated like:
    y = (x - E(X)) / (Var(x) + epsilon) * gamma + beta
    More info can see: https://pytorch.org/docs/stable/generated/torch.nn.GroupNorm.html 
    """
    def __call__(self, model):

        pattern_mapping_config = {
            'GroupNorm': [
                # wav2vec2-base / wav2vec2-large
                {
                    'patterns': {
                        'in': [[(0, 'Reshape'), (1, 'InstanceNormalization'), (3, 'Reshape'), (4, 'Mul'),
                                (5, 'Add')], [(), (2, 'Shape'), (3, 'Reshape')]],
                        'out': [[(0, 'GroupNorm')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 5
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            4: [1]
                        }, {
                            5: [1]
                        }], [[0, 1, 2], 3]]
                    },
                    'output_tensors': {
                        0: [[{
                            5: [0]
                        }], [[0], 1]]
                    },
                    'returns': [1, 4]
                }
            ]
        }

        def _set_attr(group, channels, epsilon, node_names, model):
            attr = OrderedDict()
            attr['group'] = int(group)
            attr['channels'] = int(channels)
            attr['epsilon'] = float(epsilon)
            gn_node_idx = model.get_node_id(node_names[0])
            model.nodes[gn_node_idx].attr = attr

        for i in range(len(pattern_mapping_config['GroupNorm'])):
            pattern_dict = pattern_mapping_config['GroupNorm'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("GroupNorm", pattern_dict, model)
            if len(new_node_names) != 0:
                logger.info('GroupNorm mathched...')
                logger.debug('GroupNorm = {}'.format(new_node_names))
                for j in range(len(new_node_names)):
                    group = ret_old_nodes[j][0].input_tensors[1].shape[0]
                    channels = ret_old_nodes[j][1].input_tensors[1].shape[0]
                    epsilon = 9.999999960041972e-13
                    #epsilon = ret_old_nodes[j][0].attr['epsilon']
                    _set_attr(group, channels, epsilon, new_node_names[j], model)

                return model

        return model


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry
from .tensor import Tensor
from ..graph_utils import list2str


# This operation creates a tensor of shape dims and fills it with value.
# tf.fill(dims, value, name=None)
@operator_registry(operator_type='Squeeze')
class Squeeze(Operator):
    """Parse the Squeeze operator to the neural engine."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr from tensorflow."""
        if framework == 'tensorflow':
            self._attr['squeeze_dims'] = node.attr['squeeze_dims'].list.i
        """Extract the node attr from onnxruntime."""
        if framework == 'onnxruntime':
            # ai.onnx v11
            if len(node.attribute):
                axis = node.attribute[0].ints
                if len(axis) == 1:
                    self._attr['axis'] = axis[0]
                else:
                    self._attr['axis'] = list2str(axis)
            # ai.onnx v14
            else:
                if len(self._input_tensors[1].data) == 1:
                   self._attr['axis'] = int(self._input_tensors[1].data)
                else:
                   self._attr['axis'] = list2str(self._input_tensors[1].data)
                self._input_tensors.pop()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The AttentionOutputLayerNormLengthAdaptiveExpandIndices Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
import numpy as np


@pattern_registry(pattern_type='AttentionOutputLayerNormLengthAdaptiveExpandIndices')
class AttentionOutputLayerNormLengthAdaptiveExpandIndices(Pattern):
    """The AttentionOutputLayerNormLengthAdaptiveExpandIndices pattern.

    Fuse the original sub-graph into the custom acceleration graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """

    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'AttentionOutputLayerNormLengthAdaptiveExpandIndices': [
                # minilmv2-lat-roberta-2-seq
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (6, 'Concat'),
                                (7, 'Reshape'), (8, 'Shape'), (9, 'ConstantOfShape'), (10, 'Mul'),
                                (11, 'Equal'), (12, 'Where'), (14, 'Expand')],
                               [(), (3, 'Shape'), (4, 'Gather'), (5, 'Unsqueeze'), (6, 'Concat')],
                               [(), (13, 'Unsqueeze'), (14, 'Expand')]],
                        'out': [[(0, 'ExpandIndices')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 14,
                    },
                    'input_tensors': {
                        0: [[{
                            13: [0]
                        }, {
                            0: [0]
                        }], [[0, 1], 2]],
                    },
                    'output_tensors': {
                        0: [[{
                            14: [0]
                        }], [[0], 1]],
                    },
                    'returns': [13]
                },
                # minilmv2-lat-roberta-1-seq
                # concat 155
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (3, 'Concat'),
                                (4, 'Reshape'), (5, 'Shape'), (6, 'ConstantOfShape'), (7, 'Mul'),
                                (8, 'Equal'), (9, 'Where'), (10, 'Expand')],
                               [(), (11, 'Unsqueeze'), (10, 'Expand')]],
                        'out': [[(0, 'ExpandIndices')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 10,
                    },
                    'input_tensors': {
                        0: [[{
                            11: [0]
                        }, {
                            0: [0]
                        }], [[0, 1], 2]],
                    },
                    'output_tensors': {
                        0: [[{
                            10: [0]
                        }], [[0], 1]],
                    },
                    'returns': [11]
                },
                # ConstantOfShape_573,323
                {
                    'patterns': {
                        'in': [[(0, 'ConstantOfShape'), (1, 'Mul'),
                                (2, 'Equal'), (3, 'Where'), (4, 'Expand'), (5, 'GatherElements')],
                               [(), (6, 'Unsqueeze'), (7, 'Unsqueeze'), (4, 'Expand')]],
                        'out': [[(0, 'ExpandIndices'), (1, 'GatherElements')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 4,
                        1: 5,
                    },
                    'input_tensors': {
                        0: [[{
                            6: [0]
                        }, {
                            5: [0]
                        }], [[0, 1], 2]],

                        1: [[{
                            5: [0]
                        },

                        ], [[0], 2]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            5: [0]
                        }], [[0], 1]],
                    },
                    'returns': [5,6,7]
                },
                # ConstantOfShape_272, 517
                {
                    'patterns': {
                        'in': [[(0, 'ConstantOfShape'), (1, 'Mul'),
                                (2, 'Equal'), (3, 'Where'), (4, 'Expand'), (5, 'GatherElements')],
                               [(), (6, 'Unsqueeze'), (4, 'Expand')]],
                        'out': [[(0, 'ExpandIndices'), (1, 'GatherElements')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 4,
                        1: 5,
                    },
                    'input_tensors': {
                        0: [[{
                            6: [0]
                        }, {
                            5: [0]
                        }], [[0, 1], 2]],

                        1: [[{
                            5: [0]
                        },

                        ], [[0], 2]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            5: [0]
                        }], [[0], 1]],
                    },
                    'returns': [5,6]
                },
                
                # int8 lat
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (3, 'Concat'),
                                (4, 'Reshape'), (5, 'Equal'), (6, 'Where'), (8, 'Expand')],
                               [(), (7, 'Unsqueeze'), (8, 'Expand')]],
                        'out': [[(0, 'ExpandIndices')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 7,
                    },
                    'input_tensors': {
                        0: [[{
                            7: [0]
                        }, {
                            0: [0]
                        }], [[0, 1], 2]],
                    },
                    'output_tensors': {
                        0: [[{
                            8: [0]
                        }], [[0], 1]],
                    },
                    'returns': [7]
                },
            ]
        }

        # minilmv2-lat-roberta
        #for _, pattern_dict in pattern_mapping_config.items():
        for idx in range(len(pattern_mapping_config[
            'AttentionOutputLayerNormLengthAdaptiveExpandIndices'])):

            pattern_dict = pattern_mapping_config[
                'AttentionOutputLayerNormLengthAdaptiveExpandIndices'][idx]
            model, new_node_names, ret_old_nodes = util.pattern_mapping(
                 'AttentionOutputLayerNormLengthAdaptiveExpandIndices', pattern_dict, model)
            if len(new_node_names) != 0:
                for i in range(len(new_node_names)):
                    attr = OrderedDict()
                    attr_gather = OrderedDict()
                    input_indices = []
                    axis_gather = []
                    for ret_old_node in ret_old_nodes[i]:
                        if ret_old_node.op_type == 'Unsqueeze':
                           input_indices.append(int(ret_old_node.attr['axes']))
                        elif ret_old_node.op_type == 'GatherElements':
                            axis_gather.append(int(ret_old_node.attr['axis']))


                    attr['position'] = util.list2str(input_indices)
                    keep_indices_node_idx = model.get_node_id(new_node_names[i][0])
                    model.nodes[keep_indices_node_idx].attr = attr

                    # set attr for gather_elements
                    if 1 < len(new_node_names[i]):
                       attr_gather['axis'] = util.list2str(axis_gather)
                       gather_element_node_idx = model.get_node_id(new_node_names[i][1])
                       model.nodes[gather_element_node_idx].attr = attr_gather



                #return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The Transformer2Dmodel_QKVReshape Pattern."""

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
from .. import graph_utils as util
from .. import logger


@pattern_registry(pattern_type='Transformer2Dmodel_QKVReshape')
class Transformer2Dmodel_QKVReshape(Pattern):
    """The Transformer2Dmodel_QKVReshape pattern.

    Fuse the original sub-graph into the custom acceleration 'Transformer2Dmodel_QKVReshape' graph.
    The search strategy is based on the following pattern mapping configs for the stable diffusionV1-5.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            # this pattern is espciall for nodes that the length of concat equals 3.
            'Transformer2Dmodel_QKVReshape': [
                {
                    'patterns': {
                        'in': [[(0, 'Mul'), (1, 'Unsqueeze'), (4, 'Concat'), (5, 'Reshape')],
                               [(), (2, 'Unsqueeze'), (4, 'Concat')], [(), (3, 'Unsqueeze'), (4, 'Concat')]],
                        'out': [[(0, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 5,
                    },
                    'input_tensors': {
                        0: [[{
                            5: [0]
                        }], [[0], 1]],
                    },
                    'output_tensors': {
                        0: [[{
                            5: [0]
                        }], [[0], 1]]
                    },
                    'returns': [5, 0]
                },
            ]
        }

        for i in range(len(pattern_mapping_config['Transformer2Dmodel_QKVReshape'])):
            pattern_dict = pattern_mapping_config['Transformer2Dmodel_QKVReshape'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("Transformer2Dmodel_QKVReshape",
                                                                        pattern_dict, model)
            weight = False
            if len(new_node_names) != 0:
                logger.info('Transformer2Dmodel_QKVReshape mathched...')
                logger.debug('Transformer2Dmodel_QKVReshape = {}'.format(new_node_names))
                for j in range(len(new_node_names)):
                    # the first new node
                    reshape_node_idx = model.get_node_id(new_node_names[j][0])

                    mul_node = ret_old_nodes[j][1]
                    assert mul_node.op_type == 'Mul'
                    mul_value = int(mul_node.input_tensors[1].data)

                    matmul_node_name = model.nodes[reshape_node_idx].output_tensors[0].dest_op[0]
                    matmul_node = model.get_node_by_name(matmul_node_name)
                    # especially for reshape_7 followed by MatMulWithBias
                    if matmul_node.op_type == 'MatMulWithBias':
                        weight = matmul_node.input_tensors[1].data.shape[1]
                        attr = OrderedDict()
                        attr['dst_shape'] = '-1,' + str(weight)
                        model.nodes[reshape_node_idx].attr = attr
                    else:
                        gather_node = model.get_node_by_name(
                            ret_old_nodes[j][1].input_tensors[0].source_op[0])
                        shape_node = model.get_node_by_name(gather_node.input_tensors[0].source_op[0])
                        matmul_node = model.get_node_by_name(shape_node.input_tensors[0].source_op[0])
                        weight = matmul_node.input_tensors[1].data.shape[1]
                        attr = OrderedDict()
                        attr['dst_shape'] = '16,-1,' + str(int(weight / mul_value))
                        model.nodes[reshape_node_idx].attr = attr

        return model



import os
import sys
import unittest

import torch
import numpy as np
from intel_extension_for_transformers.optimization.config import WeightPruningConfig, PrunerV2
from intel_extension_for_transformers.optimization.pytorch_pruner.pruning import Pruning
import torchvision
import torch.nn as nn
from torch.utils.data import DataLoader, TensorDataset

sys.path.insert(0, './')

local_schedulers_config = [
    {
        "start_step": 0,
        "end_step": 2,
        "pruning_type": "magnitude",
        "op_names": ['layer1.*'],
        "excluded_op_names": ['layer2.*'],
        "pruning_scope": "global",
        "target_sparsity": 0.5,
        "pattern": "4x1"
    },
    {
        "start_step": 1,
        "end_step": 10,
        "target_sparsity": 0.5,
        "pruning_type": "snip_momentum",
        "pruning_frequency": 2,
        "op_names": ['layer2.*'],
        "pruning_scope": "local",
        "target_sparsity": 0.75,
        "pattern": "32x1",
        "sparsity_decay_type": "exp"
    },
    {
        "start_step": 1,
        "end_step": 10,
        "pruning_type": "snip_progressive",
        "pruning_frequency": 2,
        "op_names": ['layer2.*'],
        "pruning_scope": "local",
        "target_sparsity": 0.7,
        "pattern": "4x2",
        "sparsity_decay_type": "linear"
    }
]

fake_snip_config = WeightPruningConfig(local_schedulers_config, target_sparsity=0.9, start_step=0, \
                                       end_step=10, pruning_frequency=1, sparsity_decay_type="exp")

def build_fake_yaml_basic():
    fake_snip_yaml = """
version: 1.0

model:
  name: "bert-mini"
  framework: "pytorch"

pruning:
  approach:
    weight_compression_v2:
      start_step: 0
      end_step: 0

      pruning_op_types: ["Linear"]
      target_sparsity: 0.5
      max_sparsity_ratio_per_op: 0.5

      pruners:
        - !PrunerV2
            excluded_op_names: ["classifier", "pooler", ".*embeddings*"]
            pattern: "2:4"
            pruning_frequency: 50
            pruning_scope: "global"
            pruning_type: "snip_momentum"
            sparsity_decay_type: "linear"
    """
    with open('fake_snip.yaml', 'w', encoding="utf-8") as f:
        f.write(fake_snip_yaml)

class TestPruning(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.model = torchvision.models.resnet18()
        build_fake_yaml_basic()

    @classmethod
    def tearDownClass(self):
        os.remove('fake_snip.yaml')

    def test_pruning_basic(self):
        local_configs = [
            {
                "op_names": ['layer1.*'],
                'target_sparsity': 0.5,
                "pattern": '8x2',
                "pruning_type": "magnitude_progressive",
                "false_key": "this is to test unsupport keys"
            },
            {
                "op_names": ['layer2.*'],
                'target_sparsity': 0.5,
                'pattern': '2:4'
            },
            {
                "op_names": ['layer3.*'],
                'target_sparsity': 0.7,
                'pattern': '5x1',
                "pruning_type": "snip_progressive"
            },
            {
                "start_step": 2,
                "end_step": 8,
                "pruning_type": "gradient",
                "pruning_frequency": 2,
                "op_names": ['fc'],
                "pruning_scope": "local",
                "target_sparsity": 0.75,
                "pattern": "1x1",
                "sparsity_decay_type": "cube",
                "reg_type": "group_lasso",
                "parameters": {'reg_coeff': 0.0}
            }
        ]
        config = WeightPruningConfig(
            local_configs,
            target_sparsity=0.8
        )
        prune = Pruning(config)
        prune.update_config(start_step=1, end_step=10)
        prune.model = self.model

        criterion = nn.CrossEntropyLoss()
        optimizer = torch.optim.SGD(self.model.parameters(), lr=0.0001)
        x_train = np.random.uniform(low=0., high=1., size=tuple([10, 3, 224, 224]))
        y_train = np.random.randint(low=0, high=2, size=tuple([10]))
        x_train, y_train = torch.tensor(x_train, dtype=torch.float32), \
                                torch.tensor(y_train, dtype=torch.long)
        dummy_dataset = TensorDataset(x_train, y_train)
        dummy_dataloader = DataLoader(dummy_dataset)

        prune.on_train_begin()
        prune.update_config(pruning_frequency=4)
        for epoch in range(2):
            self.model.train()
            prune.on_epoch_begin(epoch)
            local_step = 0
            for image, target in dummy_dataloader:
                prune.on_step_begin(local_step)
                output = self.model(image)
                loss = criterion(output, target)
                optimizer.zero_grad()
                loss.backward()
                prune.on_before_optimizer_step()
                optimizer.step()
                prune.on_after_optimizer_step()
                prune.on_step_end()
                local_step += 1

            prune.on_epoch_end()
        prune.get_sparsity_ratio()
        prune.on_train_end()
        prune.on_before_eval()
        prune.on_after_eval()

    def test_pruning_pattern(self):
        local_configs = [
            {
                "op_names": ['layer1.*'],
                'target_sparsity': 0.5,
                "pattern": '5:8',
                "pruning_type": "magnitude"
            },
            {
                "op_names": ['layer2.*'],
                "pattern": '1xchannel',
                "pruning_scope": "global"
            },
            {
                "start_step": 2,
                "end_step": 20,
                "op_names": ['layer3.*'],
                'target_sparsity': 0.666666,
                'pattern': '4x2',
                "pruning_type": "snip_progressive",
                "pruning_frequency": 5
            }
        ]
        config = WeightPruningConfig(
            local_configs,
            target_sparsity=0.8,
            sparsity_decay_type="cos",
            excluded_op_names=["downsample.*"],
            pruning_scope="local",
            min_sparsity_ratio_per_op=0.1
        )
        prune = Pruning(config)
        prune.update_config(start_step=1, end_step=10)
        prune.model = self.model

        criterion = nn.CrossEntropyLoss()
        optimizer = torch.optim.SGD(self.model.parameters(), lr=0.0001)
        x_train = np.random.uniform(low=0., high=1., size=tuple([10, 3, 224, 224]))
        y_train = np.random.randint(low=0, high=2, size=tuple([10]))
        x_train, y_train = torch.tensor(x_train, dtype=torch.float32), \
                                torch.tensor(y_train, dtype=torch.long)
        dummy_dataset = TensorDataset(x_train, y_train)
        dummy_dataloader = DataLoader(dummy_dataset)
        prune.on_train_begin()
        for epoch in range(5):
            self.model.train()
            prune.on_epoch_begin(epoch)
            local_step = 0
            for image, target in dummy_dataloader:
                prune.on_step_begin(local_step)
                output = self.model(image)
                loss = criterion(output, target)
                optimizer.zero_grad()
                loss.backward()
                prune.on_before_optimizer_step()
                optimizer.step()
                prune.on_after_optimizer_step()
                prune.on_step_end()
                local_step += 1

            prune.on_epoch_end()
        prune.get_sparsity_ratio()
        prune.on_train_end()
        prune.on_before_eval()
        prune.on_after_eval()


    def test_fucntion(self):
        from intel_extension_for_transformers.optimization.pytorch_pruner.patterns import BasePattern, PatternNInM, PatternNxM
        from intel_extension_for_transformers.optimization.pytorch_pruner.utils import process_and_check_config, process_config, parse_to_prune
        from intel_extension_for_transformers.optimization.pytorch_pruner.pruners import parse_valid_pruner_types

        prune = Pruning(fake_snip_config)
        prune.model = self.model
        prune.on_train_begin()
        config = process_and_check_config(fake_snip_config)[0]
        modules = parse_to_prune(config, self.model)

        test_pattern = BasePattern(config, modules)
        mask = test_pattern.get_single_mask_per_target_ratio(torch.randn(5, 6), 0.8)
        test_pattern.get_sparsity_ratio({'': mask})
        test_pattern.get_pattern_lock_masks(prune.pruners[0].modules)
        test_pattern.check_layer_validity()

        test_pattern = PatternNxM(config, modules)
        test_pattern.get_sparsity_ratio_progressive({'': mask})
        test_pattern.get_pattern_lock_masks(prune.pruners[0].modules)

        config = process_config('fake_snip.yaml')[0]
        test_pattern = PatternNInM(config, modules)
        test_pattern.get_pattern_lock_masks(prune.pruners[0].modules)

        parse_valid_pruner_types()


if __name__ == "__main__":
    unittest.main()



import os
import sys
import torch
import unittest
from unittest.mock import patch
from intel_extension_for_transformers.optimization.model import OptimizedModel

os.environ["CUDA_VISIBLE_DEVICES"] = ""

# example test for question-answering quantization with IPEX only for now
SRC_DIRS = [
    os.path.join("../examples/optimization/pytorch/huggingface/", dirname)
    for dirname in [
        "question-answering/quantization/inc/",
    ]
]
sys.path.extend(SRC_DIRS)

if SRC_DIRS is not None:
    import run_qa

class TestExamples(unittest.TestCase):
    def test_run_qa_ipex(self):
        test_args = f"""
            run_qa.py
            --model_name_or_path bert-large-uncased-whole-word-masking-finetuned-squad
            --dataset_name squad
            --tune
            --quantization_approach PostTrainingStatic
            --do_train
            --do_eval
            --max_eval_samples 100
            --max_train_samples 50
            --output_dir ./tmp/squad_output
            --overwrite_output_dir
            --framework ipex
            """.split()

        with patch.object(sys, "argv", test_args):
            run_qa.main()
            int8_model = OptimizedModel.from_pretrained("./tmp/squad_output")
            self.assertTrue(isinstance(int8_model, torch.jit.ScriptModule))
        
        test_args = f"""
            run_qa.py
            --model_name_or_path bert-large-uncased-whole-word-masking-finetuned-squad
            --dataset_name squad
            --quantization_approach PostTrainingStatic
            --do_train
            --do_eval
            --max_eval_samples 100
            --max_train_samples 50
            --output_dir ./tmp/squad_output
            --overwrite_output_dir
            --framework ipex
            --benchmark_only
            --cores_per_instance 16
            --num_of_instance 1
            """.split()

        with patch.object(sys, "argv", test_args):
            run_qa.main()


if __name__ == "__main__":
    unittest.main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry
from .tensor import Tensor


@operator_registry(operator_type='QuantizeLinear')
class QuantizeLinear(Operator):
    """Parse the QuantizeLinear operator to the neural engine."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr from onnxruntime."""
        self._op_type = 'Quantize'
        if framework == 'onnxruntime':
            self._attr['output_dtype'] = 'u8'
            self._attr['quant_mode'] = 'zp_scale'


@operator_registry(operator_type='Quantize')
class Quantize(Operator):
    """Register the Quantize operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr from onnxruntime."""
        # aten::quantize_per_tensor(%weight, %scale, %zero_point, %dtype)
        self._attr['scale'] = node.inputsAt(1).toIValue()
        self._attr['zero_point'] = node.inputsAt(2).toIValue()
        self._attr['dtype'] = node.inputsAt(3).toIValue()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The TextEncoder_AttentionReshape Pattern."""

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
from .. import graph_utils as util
from .. import logger


@pattern_registry(pattern_type='TextEncoder_AttentionReshape')
class TextEncoder_AttentionReshape(Pattern):
    """The TextEncoder_AttentionReshape pattern.

    Fuse the original sub-graph into the custom acceleration 'TextEncoder_AttentionReshape' graph.
    The search strategy is based on the following pattern mapping configs for the stable diffusionV1-5.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            # for text encoder self_attn/out_proj/Add
            'TextEncoder_AttentionReshape': [
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (9, 'Concat'), (10, 'Reshape'),
                                (11, 'MatMulWithBias')],
                               [(), (3, 'Shape'), (4, 'Gather'), (5, 'Unsqueeze'), (9, 'Concat')],
                               [(), (6, 'Shape'), (7, 'Gather'), (8, 'Unsqueeze'), (9, 'Concat')]],
                        'out': [[(0, 'Reshape'), (1, 'Reshape'), (1, 'MatMulWithBias')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 10,
                        1: 'TextEncoder_AttentionReshape/reshape_to_2D',
                        2: 11
                    },
                    'input_tensors': {
                        0: [[{
                            10: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                        1: [[{
                            'input_data': [0]
                        }], [[1], 2]],
                        2: [[{
                            11: [1]
                        }, {
                            11: [2]
                        }], [[1, 2], 3]],
                    },
                    'output_tensors': {
                        0: [[{
                            10: [0]
                        }], [[0], 1]],
                        1: [[], [[], 1]],
                        2: [[{
                            11: [0]
                        }], [[0], 1]]
                    },
                    'returns': [9, 11]
                },
            ]
        }

        def _set_attr(node_names, model):
            attr = OrderedDict()
            # bsz, max_seq, hidden_size, 1, 77, 768
            attr['dst_shape'] = '-1,-1,-1'
            attr['dims'] = '0, 1'
            reshape_node_idx = model.get_node_id(node_names[0])
            model.nodes[reshape_node_idx].attr = attr

            attr_1 = OrderedDict()
            # bsz, max_seq, hidden_size, 1, 77, 768
            attr_1['dst_shape'] = '-1,-1'
            attr_1['dims'] = 1
            reshape_node_idx = model.get_node_id(node_names[1])
            model.nodes[reshape_node_idx].attr = attr_1

        for i in range(len(pattern_mapping_config['TextEncoder_AttentionReshape'])):
            pattern_dict = pattern_mapping_config['TextEncoder_AttentionReshape'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("TextEncoder_AttentionReshape",
                                                                        pattern_dict, model)

            if len(new_node_names) != 0:
                logger.info('TextEncoder_AttentionReshape mathched...')
                logger.debug('TextEncoder_AttentionReshape = {}'.format(new_node_names))
                for j in range(len(new_node_names)):
                    if len(ret_old_nodes[j]) == 2:
                        _set_attr(new_node_names[j], model)

                        assert ret_old_nodes[j][1].op_type == 'MatMulWithBias'
                        mat_node_idx = model.get_node_id(new_node_names[j][2])
                        model.nodes[mat_node_idx].attr = ret_old_nodes[j][1].attr

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The PaddingSequence pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
import copy
from .. import graph_utils as util
from ..ops.tensor import Tensor


@pattern_registry(pattern_type='PaddingSequence')
class PaddingSequence(Pattern):
    """The PaddingSequence pattern.

    Fuse the original sub-graph into the custom acceleration 'PaddingSequence' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'PaddingSequence': [
                {
                    'patterns': {
                        'in': [[(1, 'Shape'), (2, 'StridedSlice'), (5, 'Pack'), (6, 'Reshape'),
                                (7, 'ExpandDims'), (8, 'Sub'), (9, 'Mul'), (10, ['Add', 'AddV2'])],
                               [(), (3, 'Shape'), (4, 'StridedSlice'), (5, 'Pack')],
                               [(), (0, 'Cast'), (6, 'Reshape')]],
                        'out': [[(0, 'AddV2')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 10
                    },
                    'input_tensors': {
                        0: [[{
                            10: [0]
                        }, {
                            'padding_sequence': [0]
                        }], [[0, 1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            10: [0]
                        }], [[0], 1]]
                    },
                    'returns': []
                },

                # bert_base_mrpc
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'StridedSlice'), (2, 'Pack'), (3, 'Fill'),
                                (7, 'Mul'), (8, 'ExpandDims'), (9, 'Sub'), (10, 'Mul'),
                                (11, ['Add', 'AddV2'])],
                               [(1, 'StridedSlice'), (4, 'Pack'), (5, 'Reshape'), (6, 'Cast'),
                                (7, 'Mul')]],
                        'out': [[(0, 'AddV2')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 11
                    },
                    'input_tensors': {
                        0: [[{
                            11: [0]
                        }, {
                            'padding_sequence': [0]
                        }], [[0, 1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            11: [0]
                        }], [[0], 1]]
                    },
                    'returns': []
                },
                # geminet, bert_related in huggingface
                {
                    'patterns': {
                        'in': [[(0, 'Unsqueeze'), (1, 'Unsqueeze'), (2, 'Cast'), (3, 'Sub'),
                                (4, 'Mul')]],
                        'out': [[(0, 'PaddingSequence')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 4
                    },
                    'input_tensors': {
                        0: [[{
                            'input_data': [-1]
                        }], [[0], 1]]
                    },
                    'output_tensors': {
                        0: [[{
                            4: [0]
                        }], [[0], 1]]
                    },
                    'returns': []
                },
                # distil_bert_base
                {
                    'patterns': {
                        'in': [[(1, 'Unsqueeze'), (3, 'Concat'), (4, 'Reshape'), (6, 'Expand'),
                                (7, 'Cast'), (8, 'Where')], [(), (2, 'Unsqueeze'), (3, 'Concat')],
                               [(), (5, 'Shape'), (6, 'Expand')], [(), (0, 'Equal'),
                                                                   (4, 'Reshape')]],
                        'out': [[(0, 'AddV2')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 8
                    },
                    'input_tensors': {
                        0: [[{
                            5: [0]
                        }, {
                            'padding_sequence': [0]
                        }], [[0, 1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            8: [0]
                        }], [[0], 1]]
                    },
                    'returns': [2]
                },

                # distil_bert_base_int8
                {
                    'patterns': {
                        'in': [[(0, 'Unsqueeze'), (1, 'Concat'), (2, 'Reshape'), (3, 'Expand'),
                                (4, 'Where')], [(), (5, 'Unsqueeze'), (1, 'Concat')],
                               [(), (6, 'Shape'), (3, 'Expand')], [(), (7, 'Equal'),
                                                                   (2, 'Reshape')]],
                        'out': [[(0, 'AddV2')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 4
                    },
                    'input_tensors': {
                        0: [[{
                            6: [0]
                        }, {
                            'padding_sequence': [0]
                        }], [[0, 1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            4: [0]
                        }], [[0], 1]]
                    },
                    'returns': [5]
                },

                # opennmt encoder
                {
                    'patterns': {
                        'in': [[(0, 'Input'), (1, 'ReduceMax'), (2, 'Cast'), (3, 'Range'),
                                (5, 'Expand'), (6, 'Tile'), (8, 'Less'), (9, 'Unsqueeze'),
                                (10, 'Not'), (11, 'Unsqueeze'), (12, 'Cast'), (13, 'Where')],
                               [(), (4, 'ConstantOfShape'), (5, 'Expand')],
                               [(0, 'Input'), (7, 'Unsqueeze'), (8, 'Less')]],
                        'out': [[(0, 'AddV2')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 13
                    },
                    'input_tensors': {
                        0: [[{
                            13: [2]
                        }, {
                            'padding_sequence': [0]
                        }], [[0, 1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            13: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0]
                },

                # opennmt decoder
                {
                    'patterns': {
                        'in': [[(0, 'Input'), (1, 'Shape'), (2, 'Gather'), (3, 'Cast'),
                                (4, 'Range'), (6, 'Expand'), (7, 'Tile'), (9, 'Less'),
                                (10, 'Unsqueeze'), (11, 'Not'), (12, 'Unsqueeze'), (13, 'Cast'),
                                (14, 'Where')],
                               [(), (5, 'ConstantOfShape'), (6, 'Expand')],
                               [(0, 'Input'), (8, 'Unsqueeze'), (9, 'Less')]],
                        'out': [[(0, 'AddV2')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 14
                    },
                    'input_tensors': {
                        0: [[{
                            14: [2]
                        }, {
                            'padding_sequence': [0]
                        }], [[0, 1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            14: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0]
                },
            ]
        }

        def _make_padding_sequence_node(tensor_idx, hidden_size, model, seq_len_first=False):
            # Models with different size may have different nums of self-attention head.
            # In Google Bert, the nums_head = hidden_size // 64.
            # But in some other models, like minilm, they has smaller attention head channel.
            # Use this dict to maintain the attr of padding_sequence operation.
            # The last key is for debug and tell the info of unknown model.
            heads_num_dict = {1024: 16, 768: 12, 384: 12, 256: 4, 128: 2, 32: 2, -1: -1}
            node_name = 'padding_sequence'
            heads_num = int(heads_num_dict[hidden_size])
            input_data = model.nodes[0]
            if len(input_data.output_tensors) > tensor_idx:
                input_tensors = [copy.deepcopy(input_data.output_tensors[tensor_idx])]
            else:
                input_tensors = [Tensor()]
            output_tensors = [Tensor(name=node_name + ':0', source_op=[node_name], dest_op=[])]
            attr = OrderedDict()
            if not is_lat_model(model):
                attr['dst_shape'] = '-1,' + str(heads_num) + ',0,-1'
            else:
                attr['dst_shape'] = '-1,' + str(1) + ',1,-1'

            if not seq_len_first:
                attr['dims'] = 1
            else:
                attr['dims'] = 0
                attr['seq_len_first'] = True
            padding_sequence_node = util.construct_node(node_name,
                                                        'PaddingSequence',
                                                        input_tensors=input_tensors,
                                                        output_tensors=output_tensors,
                                                        attr=attr)

            model.insert_nodes(1, [padding_sequence_node])

            return model

        def get_hidden_size(model, p=None, mat_idx=0):
            if p == None:
                p = [[(0, 'MatMul'), (1, ['Add', 'AddV2']), (2, ['Add', 'AddV2']),
                      (3, 'LayerNorm')]]
            match_result = util.search_pattern(p, model)
            if len(match_result) != 0:
                mat_node = model.get_node_by_name(match_result[0][mat_idx])
                hidden_size = int(mat_node.input_tensors[1].shape[-1])
            else:
                hidden_size = -1
            return hidden_size

        def is_lat_model(model, p=None):
            if p == None:
                p = [[(0, 'TopK'),(1, 'Unsqueeze'),(2, 'Unsqueeze'),(3, 'Expand'),
                    (4, 'GatherElements')]]
            match_result = util.search_pattern(p, model)
            return len(match_result) != 0

        pattern_dict = pattern_mapping_config['PaddingSequence'][0]
        model = _make_padding_sequence_node(2, 1024, model)
        model, new_node_names, ret_old_nodes = util.pattern_mapping("PaddingSequence",
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            return model
        else:
            model.remove_nodes(['padding_sequence'])

        pattern_dict = pattern_mapping_config['PaddingSequence'][1]
        model = _make_padding_sequence_node(2, 768, model)
        model, new_node_names, ret_old_nodes = util.pattern_mapping("PaddingSequence",
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            return model
        else:
            model.remove_nodes(['padding_sequence'])

        pattern_dict = pattern_mapping_config['PaddingSequence'][2]
        hidden_size = get_hidden_size(model)
        if len(model.nodes[0].output_tensors) == 3:
            model = _make_padding_sequence_node(2, hidden_size, model)
        else:
            model = _make_padding_sequence_node(1, hidden_size, model)
        model, new_node_names, ret_old_nodes = util.pattern_mapping("PaddingSequence",
                                                                    pattern_dict, model)

        if len(new_node_names) != 0:
            assert hidden_size!=-1, "Wrong hidden size in padding_sequence!"
            ps_attr = model.get_node_by_name('padding_sequence').attr
            for j in range(len(new_node_names)):
                ps_node_id = model.get_node_id(new_node_names[j][0])
                model.nodes[ps_node_id].attr = ps_attr
            # remove fake node
            model.remove_nodes(['padding_sequence'])
            return model
        else:
            model.remove_nodes(['padding_sequence'])

        pattern_dict = pattern_mapping_config['PaddingSequence'][3]
        model = _make_padding_sequence_node(1, 768, model)
        model, new_node_names, ret_old_nodes = util.pattern_mapping("PaddingSequence",
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            # remove shape+gather in distil_bert_base
            for i in range(len(ret_old_nodes)):
                gather_node_name = ret_old_nodes[i][0].input_tensors[0].source_op[0]
                gather_node = model.get_node_by_name(gather_node_name)
                shape_node_name = gather_node.input_tensors[0].source_op[0]
                model.remove_nodes([gather_node_name, shape_node_name])
            return model
        else:
            model.remove_nodes(['padding_sequence'])

        pattern_dict = pattern_mapping_config['PaddingSequence'][4]
        model = _make_padding_sequence_node(1, 768, model)
        model, new_node_names, ret_old_nodes = util.pattern_mapping("PaddingSequence",
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            # remove gather in distil_bert_base
            for i in range(len(ret_old_nodes)):
                gather_node_name = ret_old_nodes[i][0].input_tensors[0].source_op[0]
                model.remove_nodes([gather_node_name])
            return model
        else:
            model.remove_nodes(['padding_sequence'])

        for idx in [5, 6]:
            pattern_dict = pattern_mapping_config['PaddingSequence'][idx]
            model = _make_padding_sequence_node(0, 1024, model, seq_len_first=True)
            model, new_node_names, ret_old_nodes = util.pattern_mapping(
                "PaddingSequence", pattern_dict, model)
            if len(new_node_names) != 0:
                assert ret_old_nodes[0][0].op_type == 'Input'
                model.insert_nodes(0, [ret_old_nodes[0][0]])
                return model
            else:
                model.remove_nodes(['padding_sequence'])

        return model



"""Basic NAS approach class."""

#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from intel_extension_for_transformers.optimization.config import NASConfig
from neural_compressor.conf.config import Conf
from neural_compressor.experimental.nas.nas import NASBase
from neural_compressor.experimental.nas.nas_utils import nas_registry

@nas_registry("Basic")
class BasicNAS(NASBase):
    """Basic NAS approach.

    Defining the pipeline for basic NAS approach.

    Args:
        conf_fname (string or obj): The path to the YAML configuration file or the object of NASConfig.
        search_space (dict): A dictionary for defining the search space.
        model_builder (function obj): A function to build model instance with the specified
            model architecture parameters.
    """

    def __init__(self, conf_fname_or_obj, search_space=None, model_builder=None):
        """Initialize the attributes."""
        NASBase.__init__(self, search_space=search_space, model_builder=model_builder)
        self._train_func = None
        self._eval_func = None
        self.init_by_cfg(conf_fname_or_obj)

    def execute(self, res_save_path=None):
        """Execute the search process.

        Returns:
            Best model architectures found in the search process.
        """
        return self.search(res_save_path)

    def estimate(self, model):
        """Estimate performance of the model.

        Depends on specific NAS algorithm. Here we use train and evaluate.

        Returns:
            Evaluated metrics of the model.
        """
        assert self._train_func is not None and self._eval_func is not None, \
            "train_func and eval_func must be set."
        self._train_func(model)
        return self._eval_func(model)

    def init_by_cfg(self, conf_fname_or_obj):
        """Initialize the configuration."""
        if isinstance(conf_fname_or_obj, str):
            if os.path.isfile(conf_fname_or_obj):
                self.conf = Conf(conf_fname_or_obj).usr_cfg
            else: # pragma: no cover
                raise FileNotFoundError(
                    "{} is not a file, please provide a NAS config file path.".format(
                        conf_fname_or_obj
                    )
                )
        elif isinstance(conf_fname_or_obj, NASConfig):
            self.config = conf_fname_or_obj.config
        else: # pragma: no cover
            raise NotImplementedError(
                "Please provide a str path to the config file or an object of NASConfig."
            )
        assert self.config.nas is not None, "nas section must be set"
        # search related config
        self.init_search_cfg(self.config.nas)

    @property
    def train_func(self): # pragma: no cover
        """Not support get train_func."""
        assert False, 'Should not try to get the value of `train_func` attribute.'

    @train_func.setter
    def train_func(self, user_train_func):
        """Training function.

        Args:
            user_train_func: This function takes "model" as input parameter
                         and executes entire training process with self
                         contained training hyper-parameters. If training_func set,
                         an evaluation process must be triggered and user should
                         set eval_dataloader with metric configured or directly eval_func
                         to make evaluation of the model executed. training_func will return
                         a trained model.
        """
        self._train_func = user_train_func

    @property
    def eval_func(self): # pragma: no cover
        """Not support get eval_func."""
        assert False, 'Should not try to get the value of `eval_func` attribute.'

    @eval_func.setter
    def eval_func(self, user_eval_func):
        """Eval function for component.

        Args:
            user_eval_func: This function takes "model" as input parameter
                         and executes entire evaluation process with self
                         contained metrics. If eval_func set,
                         an evaluation process must be triggered
                         to make evaluation of the model executed.
        """
        self._eval_func = user_eval_func

    def __repr__(self):
        """Class representation."""
        return 'BasicNAS' # pragma: no cover


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The InputData Pattern."""

import copy
from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util


@pattern_registry(pattern_type='InputData')
class InputData(Pattern):
    """The InputData pattern.

    Fuse the original sub-graph into the custom acceleration 'InputData' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'InputData': [
                {
                    'patterns': {
                        'in': [[(0, 'input_ids'), (1, 'segment_ids'), (2, 'input_mask')]],
                        'out': [[(0, 'Input')]]
                    },
                    'search_mode': 'node_name',
                    'node_names': {
                        0: 'input_data'
                    },
                    'input_tensors': {
                        0: [[], [[], 0]]
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            1: [0]
                        }, {
                            2: [0]
                        }], [[0, 1, 2], 3]]
                    },
                    'returns': [0, 1, 2]
                },
                
                # onnx model from huggingface
                {
                    'patterns': {
                        'in': [[(0, 'input_ids'), (1, 'token_type_ids'), (2, 'attention_mask')]],
                        'out': [[(0, 'Input')]]
                    },
                    'search_mode': 'node_name',
                    'node_names': {
                        0: 'input_data'
                    },
                    'input_tensors': {
                        0: [[], [[], 0]]
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            1: [0]
                        }, {
                            2: [0]
                        }], [[0, 1, 2], 3]]
                    },
                    'returns': [0, 1, 2]
                },

                # bert_base
                {
                    'patterns': {
                        'in': [[(0, 'IteratorGetNext')]],
                        'out': [[(0, 'Input')]]
                    },
                    'search_mode': 'node_name',
                    'node_names': {
                        0: 'input_data'
                    },
                    'input_tensors': {
                        0: [[], [[], 0]]
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            0: [4]
                        }, {
                            0: [1]
                        }], [[0, 1, 2], 3]]
                    },
                    'returns': []
                },

                # dlrm
                {
                    'patterns': {
                        'in': [[(0, 'dense_x'), (1, 'offsets'), (2, 'indices')]],
                        'out': [[(0, 'Input')]]
                    },
                    'search_mode': 'node_name',
                    'node_names': {
                        0: 'input_data'
                    },
                    'input_tensors': {
                        0: [[], [[], 0]]
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            1: [0]
                        }, {
                            2: [0]
                        }], 
                        [[0, 1, 2], 3]]
                    },
                    'returns': [0, 1, 2]
                },

                # distil_bert_base
                {
                    'patterns': {
                        'in': [[(0, 'input_ids'), (1, 'attention_mask') ]],
                        'out': [[(0, 'Input')]]
                    },
                    'search_mode': 'node_name',
                    'node_names': {
                        0: 'input_data'
                    },
                    'input_tensors': {
                        0: [[], [[], 0]]
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            1: [0]
                        }
                        ], [[0, 1], 2]]
                    },
                    'returns': []
                },

                # stale diffusion text encdoer
                {
                    'patterns': {
                        'in': [[(0, 'input_ids')]],
                        'out': [[(0, 'Input')]]
                    },
                    'search_mode': 'node_name',
                    'node_names': {
                        0: 'input_data'
                    },
                    'input_tensors': {
                        0: [[], [[], 0]]
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }], [[], 1]]
                    },
                    'returns': []
                },

                # stale diffusion unet
                {
                    'patterns': {
                        'in': [[(0, 'sample'), (1, 'timestep'), (2, 'encoder_hidden_states')]],
                        'out': [[(0, 'Input')]]
                    },
                    'search_mode': 'node_name',
                    'node_names': {
                        0: 'input_data'
                    },
                    'input_tensors': {
                        0: [[], [[], 0]]
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            1: [0]
                        }, {
                            2: [0]
                        }], 
                        [[0, 1, 2], 3]]
                    },
                    'returns': [0, 1, 2]
                },

                # stale diffusion vae decoder
                {
                    'patterns': {
                        'in': [[(0, 'latent_sample')]],
                        'out': [[(0, 'Input')]]
                    },
                    'search_mode': 'node_name',
                    'node_names': {
                        0: 'input_data'
                    },
                    'input_tensors': {
                        0: [[], [[], 0]]
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }], [[], 1]]
                    },
                    'returns': []
                },
                
                # minilmv2-lat-roberta
                {
                    'patterns': {
                        'in': [[(0, 'input_ids'), (1, 'input_mask') ]],
                        'out': [[(0, 'Input')]]
                    },
                    'search_mode': 'node_name',
                    'node_names': {
                        0: 'input_data'
                    },
                    'input_tensors': {
                        0: [[], [[], 0]]
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            1: [0]
                        }
                        ], [[0, 1], 2]]
                    },
                    'returns': []
                },

                # vit
                {
                    'patterns': {
                        'in': [[(0, ['input', 'pixel_values'])]],
                        'out': [[(0, 'Input')]]
                    },
                    'search_mode': 'node_name',
                    'node_names': {
                        0: 'input_data'
                    },
                    'input_tensors': {
                        0: [[], [[], 0]]
                    },
                    'output_tensors': {
                        0: [[{0: [0]}], [[0], 1]]
                    },
                    'returns': [0]
                },

            ]
        }

        for i in range(len(pattern_mapping_config['InputData'])):
            pattern_dict = pattern_mapping_config['InputData'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("InputData",
                                                                        pattern_dict, model)
            if len(new_node_names) != 0:
                model.nodes[0].attr = None
                for j in range(len(model.nodes[0].output_tensors)):
                    if model.nodes[0].output_tensors[j].shape is None:
                        model.nodes[0].output_tensors[j].shape = [-1, -1]

                return model

        # if no input_data, insert input node for subgraph
        if model.nodes[0].op_type != "Input":
            onnx_input_nodes_list = []
            model_input_tensors = []
            for node in model.nodes:
                if node.op_type == "ONNXINPUT":
                    onnx_input_nodes_list.append(node.name)
                    for input_tensor in node.output_tensors:
                         model_input_tensors.append(copy.deepcopy(input_tensor))
            input_data_node = util.construct_node('input_data',
                                                   'Input',
                                                   output_tensors=model_input_tensors)
            model.insert_nodes(0, [input_data_node])
            model.nodes[0].attr = None
            model.remove_nodes(onnx_input_nodes_list)

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry
from .tensor import Tensor
import copy

# Given a tensor x and a tensor y ,
# this operation computes x^y for corresponding elements in x and y
@operator_registry(operator_type='Pow')
class Pow(Operator):
    """Register the Pow operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()
    def set_attr(self, framework, node):
        """Extract the node attr from onnxruntime."""
        if framework == 'torch':
            import numpy as np
            if node.inputsAt(1).type().kind() != 'TensorType':
                # import pdb; pdb.set_trace()
                data = np.array(node.inputsAt(1).toIValue()).astype(np.float32)
                input_tensor = Tensor(name=self.name + '_other',
                    source_op=[],
                    dest_op=[self.name],
                    shape=[1],
                    data=data,
                    dtype="fp32"
                    )
                self.input_tensors.append(input_tensor)